from pathlib import Path
from time import sleep

import caliscope.logger
from caliscope.calibration.intrinsic_calibrator import IntrinsicCalibrator
from caliscope.cameras.camera_array import CameraData
from caliscope.gui.frame_emitters.playback_frame_emitter import PlaybackFrameEmitter
from caliscope.packets import Tracker
from caliscope.recording.recorded_stream import RecordedStream
from caliscope.trackers.charuco_tracker import CharucoTracker

logger = caliscope.logger.get(__name__)


class IntrinsicStreamManager:
    def __init__(
        self,
        recording_dir: Path,
        cameras: dict[CameraData],
        tracker: Tracker = None,
    ) -> None:
        self.recording_dir = recording_dir
        self.cameras = cameras
        self.tracker = tracker
        self.load_stream_tools()

    def load_stream_tools(self):
        self.streams = {}
        self.calibrators = {}
        self.frame_emitters = {}

        for camera in self.cameras.values():
            logger.info(f"Loading stream tools associated with camera {camera.port}")
            stream = RecordedStream(
                directory=self.recording_dir,
                port=camera.port,
                rotation_count=camera.rotation_count,
                tracker=self.tracker,
                break_on_last=False,
            )

            self.streams[camera.port] = stream

            # note: must create calibrator first so that grid history q can
            # then be passed to the frame_emitter
            self.calibrators[camera.port] = IntrinsicCalibrator(camera, stream)
            self.frame_emitters[camera.port] = PlaybackFrameEmitter(
                stream, grid_history_q=self.calibrators[camera.port].grid_history_q
            )
            self.frame_emitters[camera.port].start()

    def close_stream_tools(self):
        for port, emitter in self.frame_emitters.items():
            logger.info(f"Beginning to shut down frame emitter for port {port}")
            emitter.stop()

            logger.info(f"Waiting on camera {port} emitter to wrap up...")
            emitter.wait()
            logger.info(f"Finished waiting for camera {port} emitter to wrap up")

        for port, calibrator in self.calibrators.items():
            logger.info("stopping calibrator")
            calibrator.stop()

        for port, stream in self.streams.items():
            stream.stop_event.set()
            stream.unpause()
            logger.info(f"About to wait for camera {port} to close")
            # stream.jump_to(0)
            stream.thread.join()

        logger.info("Finished closing stream tools")

    def get_frame_count(self, port):
        """
        Note that if frame_index may not start at 0 if frame_history is from a real-time capture
        """
        start_frame_index = self.streams[port].start_frame_index
        last_frame_index = self.streams[port].last_frame_index

        return last_frame_index - start_frame_index + 1

    def update_charuco(self, charuco_tracker: CharucoTracker):
        for stream in self.streams.values():
            stream.tracker = charuco_tracker

        for emitter in self.frame_emitters.values():
            emitter.initialize_grid_capture_history()

    def play_stream(self, port):
        logger.info(f"Begin playing stream at port {port}")
        self.streams[port].play_video()

    def pause_stream(self, port):
        logger.info(f"Pausing stream at port {port}")
        self.streams[port].pause()

    def unpause_stream(self, port):
        logger.info(f"Unpausing stream at port {port}")
        self.streams[port].unpause()

    def stream_jump_to(self, port, frame):
        self.streams[port].jump_to(frame)

    def end_stream(self, port):
        self.streams[port].stop_event.set()
        self.unpause_stream(port)

    def add_calibration_grid(self, port: int, frame_index: int):
        self.calibrators[port].add_calibration_frame_index(frame_index)

    def clear_calibration_data(self, port: int):
        self.calibrators[port].clear_calibration_data()
        self.frame_emitters[port].initialize_grid_capture_history()

    def calibrate_camera(self, port: int):
        logger.info(f"Calibrating camera at port {port}")
        self.calibrators[port].calibrate_camera()

    def apply_distortion(self, camera: CameraData, undistort: bool):
        self.frame_emitters[camera.port].update_distortion_params(
            undistort=undistort, matrix=camera.matrix, distortions=camera.distortions
        )

    def set_stream_rotation(self, port, rotation_count):
        self.streams[port].rotation_count = rotation_count

    def autocalibrate(self, port, grid_count, pct_board_threshold):
        stream = self.streams[port]
        intrinsic_calibrator = self.calibrators[port]
        frame_emitter = self.frame_emitters[port]

        board_corners = self.tracker.charuco.board.getChessboardCorners()
        total_corner_count = board_corners.shape[0]
        threshold_corner_count = total_corner_count * pct_board_threshold
        threshold_corner_count = max(
            threshold_corner_count, 6
        )  # additional requirement that I believe is part of the alogrithm

        logger.info(f"Corners for charuco are {board_corners}")

        # calculate basic wait time between board collections
        # if many frames have incomplete data, this will fail to reach the target board count
        start_frame_index = stream.start_frame_index
        last_frame_index = stream.last_frame_index
        total_frames = last_frame_index - start_frame_index + 1
        wait_between = int(total_frames / grid_count)

        stream.set_fps_target(100)  # speed through the stream

        # jump to first frame, play videos and cycle quickly through frames
        stream.jump_to(0)
        frame_emitter.initialize_grid_capture_history()
        intrinsic_calibrator.initiate_auto_pop(
            wait_between=wait_between,
            threshold_corner_count=threshold_corner_count,
            target_grid_count=grid_count,
        )

        stream.unpause()

        while intrinsic_calibrator.grid_count < grid_count:
            logger.info(f"Waiting for sufficient calibration boards to become populated at port {port}")
            sleep(2)

        intrinsic_calibrator.calibrate_camera()



from pathlib import Path

import rtoml

import caliscope.logger
from caliscope.tracker import Segment, WireFrameView

logger = caliscope.logger.get(__name__)


def get_wireframe(toml_spec_path: Path, point_names: dict) -> WireFrameView:
    # load in toml
    wireframe_specs = rtoml.load(toml_spec_path)

    logger.info(f"Building following wireframe: {wireframe_specs}")

    # build out a list of segments based on the dictionary
    segments = []

    for segment_name, specs in wireframe_specs.items():
        segment = Segment(
            name=segment_name,
            color=specs["color"],
            point_A=specs["points"][0],
            point_B=specs["points"][1],
        )
        segments.append(segment)

    wireframe = WireFrameView(segments=segments, point_names=point_names)

    return wireframe



from collections import OrderedDict
from pathlib import Path
from time import sleep, time

import numpy as np
from PySide6.QtCore import QObject, QThread, Signal

import caliscope.logger
from caliscope.calibration.capture_volume.capture_volume import CaptureVolume
from caliscope.calibration.capture_volume.helper_functions.get_point_estimates import (
    get_point_estimates,
)
from caliscope.calibration.capture_volume.point_estimates import PointEstimates
from caliscope.calibration.capture_volume.quality_controller import QualityController
from caliscope.calibration.charuco import Charuco
from caliscope.calibration.stereocalibrator import StereoCalibrator
from caliscope.cameras.camera_array import CameraArray, CameraData
from caliscope.cameras.camera_array_initializer import CameraArrayInitializer
from caliscope.configurator import Configurator
from caliscope.intrinsic_stream_manager import IntrinsicStreamManager
from caliscope.post_processing.post_processor import PostProcessor
from caliscope.synchronized_stream_manager import (
    SynchronizedStreamManager,
    read_video_properties,
)
from caliscope.trackers.charuco_tracker import CharucoTracker
from caliscope.trackers.tracker_enum import TrackerEnum
from caliscope.workspace_guide import WorkspaceGuide

logger = caliscope.logger.get(__name__)


FILTERED_FRACTION = (
    0.025  # by default, 2.5% of image points with highest reprojection error are filtered out during calibration
)


# class CalibrationStage(Enum):
#     NO_INTRINSIC_VIDEO = auto()
#     INTRINSIC_VIDEO_NO_INTRINSIC_CAL = auto()
#     PARTIAL_INTRINSICS = auto()
#     FULL_INTRINSICS_NO_EXTRINSICS = auto()
#     FULLY_CALIBRATED = auto()


class Controller(QObject):
    """
    Thin layer to integrate GUI and backend
    Tracks stage of the calibration based on a variety of factors
    """

    new_camera_data = Signal(int, OrderedDict)  # port, camera_display_dictionary
    capture_volume_calibrated = Signal()
    capture_volume_shifted = Signal()
    enable_inputs = Signal(int, bool)  # port, enable
    post_processing_complete = Signal()
    show_synched_frames = Signal()

    def __init__(self, workspace_dir: Path):
        super().__init__()
        self.workspace = workspace_dir
        self.config = Configurator(self.workspace)
        self.camera_count = self.config.get_camera_count()

        # streams will be used to play back recorded video with tracked markers to select frames
        self.camera_array = CameraArray({})  # empty camera array at init
        logger.info("Retrieving charuco from config")
        self.charuco = self.config.get_charuco()
        self.charuco_tracker = CharucoTracker(self.charuco)

        logger.info("Building workpace guide")
        self.workspace_guide = WorkspaceGuide(self.workspace, self.camera_count)
        self.workspace_guide.intrinsic_dir.mkdir(exist_ok=True, parents=True)
        self.workspace_guide.extrinsic_dir.mkdir(exist_ok=True, parents=True)
        self.workspace_guide.recording_dir.mkdir(exist_ok=True, parents=True)

        self.capture_volume = None

        # needs to exist before main widget can connect to its finished signal
        self.load_workspace_thread = QThread()
        self.calibrate_camera_threads = {}
        self.autocalibrate_threads = {}

    def load_workspace(self):
        def worker():
            logger.info("Assess whether to load cameras")
            if self.workspace_guide.all_instrinsic_mp4s_available():
                self.load_camera_array()
                self.load_intrinsic_stream_manager()
                self.cameras_loaded = True
            else:
                self.cameras_loaded = False

            logger.info("Assess whether to load capture volume")
            if self.all_extrinsics_estimated():
                logger.info("All extrinsics calibrated...loading capture volume")
                self.load_estimated_capture_volume()
                self.capture_volume_loaded = True
            else:
                logger.info("Not all extrinsics calibrated...not loading capture volume")
                self.capture_volume_loaded = False

        self.load_workspace_thread.run = worker
        self.load_workspace_thread.start()

    def set_camera_count(self, count: int):
        self.camera_count = count
        self.config.save_camera_count(count)

    def get_camera_count(self) -> int:
        count = self.config.get_camera_count()
        self.camera_count = count
        return count

    def all_instrinsic_mp4s_available(self) -> bool:
        return self.workspace_guide.all_instrinsic_mp4s_available()

    def all_extrinsic_mp4s_available(self) -> bool:
        return self.workspace_guide.all_extrinsic_mp4s_available()

    def all_intrinsics_estimated(self) -> bool:
        """
        At this point, processing extrinsics and calibrating capture volume should be allowed
        """
        return self.camera_array.all_intrinsics_calibrated()

    def all_extrinsics_estimated(self) -> bool:
        """
        At this point, the capture volume tab should be available
        """
        cameras_good = self.camera_array.all_extrinsics_calibrated()
        logger.info(f"All extrinsics calculated: {cameras_good}")
        point_estimates_good = self.config.point_estimates_toml_path.exists()
        logger.info(f"Point estimates available: {point_estimates_good}")
        all_data_available = self.workspace_guide.all_extrinsic_mp4s_available()
        logger.info(f"All underlying data available: {all_data_available}")

        return cameras_good and point_estimates_good and all_data_available

    def recordings_available(self) -> bool:
        return len(self.workspace_guide.valid_recording_dirs()) > 0

    def get_charuco_params(self) -> dict:
        return self.config.dict["charuco"]

    def update_charuco(self, charuco: Charuco):
        self.charuco = charuco
        self.config.save_charuco(self.charuco)
        self.charuco_tracker = CharucoTracker(self.charuco)

        if hasattr(self, "intrinsic_stream_manager"):
            logger.info("Updating charuco within the intrinsic stream manager")
            self.intrinsic_stream_manager.update_charuco(self.charuco_tracker)

    def load_extrinsic_stream_manager(self):
        logger.info(f"Loading manager for streams saved to {self.workspace_guide.extrinsic_dir}")
        self.extrinsic_stream_manager = SynchronizedStreamManager(
            recording_dir=self.workspace_guide.extrinsic_dir,
            all_camera_data=self.camera_array.cameras,
            tracker=self.charuco_tracker,
        )

    # def process_extrinsic_streams(self, fps_target=None):
    #     def worker():
    #         output_path = Path(
    #             self.workspace_guide.extrinsic_dir, "CHARUCO", "xy_CHARUCO.csv"
    #         )
    #         output_path.unlink()  # make sure this doesn't exist to begin with.

    #         self.load_extrinsic_stream_manager()
    #         self.extrinsic_stream_manager.process_streams(fps_target=fps_target)

    #         logger.info(
    #             f"Processing of extrinsic calibration begun...waiting for output to populate: {output_path}"
    #         )
    #         while not output_path.exists():
    #             sleep(0.5)
    #             logger.info(
    #                 f"Waiting for 2D tracked points to populate at {output_path}"
    #             )

    def load_intrinsic_stream_manager(self):
        self.intrinsic_stream_manager = IntrinsicStreamManager(
            recording_dir=self.workspace_guide.intrinsic_dir,
            cameras=self.camera_array.cameras,
            tracker=self.charuco_tracker,
        )
        logger.info("Intrinsic stream manager has loaded")

        # signal to main GUI that the Camera tab needs to be reloaded
        # self.intrinsicStreamsLoaded.emit()

    def load_camera_array(self):
        """
        Loads self.camera_array by first populating self.all_camera_data
        """
        # load all previously configured data if it is there
        preconfigured_cameras = self.config.get_configured_camera_data()
        self.camera_array = CameraArray(preconfigured_cameras)

        # double check that no new camera associated files have been placed in the intrinsic calibration folder
        all_ports = self.workspace_guide.get_ports_in_dir(self.workspace_guide.intrinsic_dir)

        for port in all_ports:
            if port not in self.camera_array.cameras:
                self._add_camera_from_source(port)

    def _add_camera_from_source(self, port: int):
        """
        When adding source video to calibrate a camera, the function returns the camera index
        File will be transferred to workspace/calibration/intrinsic/port_{index}.mp4
        in keeping with project layout
        """
        # copy source over to standard workspace structure
        target_mp4_path = Path(self.workspace_guide.intrinsic_dir, f"port_{port}.mp4")
        video_properties = read_video_properties(target_mp4_path)
        size = video_properties["size"]
        new_cam_data = CameraData(
            port=port,
            size=size,
        )
        self.camera_array.cameras[port] = new_cam_data
        self.config.save_camera_array(self.camera_array)

    def get_intrinsic_stream_frame_count(self, port):
        return self.intrinsic_stream_manager.get_frame_count(port)

    def play_intrinsic_stream(self, port):
        logger.info(f"Begin playing stream at port {port}")
        self.intrinsic_stream_manager.play_stream(port)

    def pause_intrinsic_stream(self, port):
        logger.info(f"Pausing stream at port {port}")
        self.intrinsic_stream_manager.pause_stream(port)

    def unpause_intrinsic_stream(self, port):
        logger.info(f"Unpausing stream at port {port}")
        self.intrinsic_stream_manager.unpause_stream(port)

    def stream_jump_to(self, port, frame_index):
        logger.info(f"Jump to frame {frame_index} at port {port}")
        self.intrinsic_stream_manager.stream_jump_to(port, frame_index)

    def end_stream(self, port):
        self.intrinsic_stream_manager.end_stream(port)

    def add_calibration_grid(self, port: int, frame_index: int):
        self.intrinsic_stream_manager.add_calibration_grid(port, frame_index)

    def clear_calibration_data(self, port: int):
        self.intrinsic_stream_manager.clear_calibration_data(port)

    def scale_intrinsic_stream(self, port, new_scale):
        self.intrinsic_stream_manager.frame_emitters[port].set_scale_factor(new_scale)

    def calibrate_camera(self, port):
        def worker():
            if self.intrinsic_stream_manager.calibrators[port].grid_count > 0:
                self.enable_inputs.emit(port, False)
                self.camera_array.cameras[port].erase_calibration_data()
                logger.info(f"Calibrating camera at port {port}")
                self.intrinsic_stream_manager.calibrate_camera(port)

                camera_data = self.camera_array.cameras[port]
                self.config.save_camera(camera_data)
                self.push_camera_data(port)
                self.enable_inputs.emit(port, True)
            else:
                logger.warning("Not enough grids available to calibrate")

        self.calibrate_camera_threads[port] = QThread()
        self.calibrate_camera_threads[port].run = worker
        self.calibrate_camera_threads[port].start()

    def push_camera_data(self, port):
        logger.info(f"Pushing camera data for port {port}")
        camera_display_data = self.camera_array.cameras[port].get_display_data()
        logger.info(f"camera display data is {camera_display_data}")
        self.new_camera_data.emit(port, camera_display_data)

    def apply_distortion(self, port, undistort: bool):
        camera = self.camera_array.cameras[port]
        self.intrinsic_stream_manager.apply_distortion(camera, undistort)

    def rotate_camera(self, port, change):
        camera_data = self.camera_array.cameras[port]
        count = camera_data.rotation_count
        count += change
        if count in [-4, 4]:
            # reset if it completes a revolution
            camera_data.rotation_count = 0
        else:
            camera_data.rotation_count = count

        # note that extrinsic streams not altered.... just reload an replay
        self.intrinsic_stream_manager.set_stream_rotation(port, camera_data.rotation_count)

        self.push_camera_data(port)
        self.config.save_camera(camera_data)

    def load_estimated_capture_volume(self):
        """
        Following capture volume optimization via bundle adjustment, or alteration
        via a transform of the origin, the entire capture volume can be reloaded
        from the config data without needing to go through the steps

        """
        logger.info("Beginning to load estimated capture volume")
        self.point_estimates = self.config.get_point_estimates()
        # self.camera_array = self.config.get_camera_array()
        self.capture_volume = CaptureVolume(self.camera_array, self.point_estimates)
        logger.info("Load of capture volume complete")

        self.capture_volume.stage = self.config.dict["capture_volume"]["stage"]
        if "origin_sync_index" in self.config.dict["capture_volume"].keys():
            self.capture_volume.origin_sync_index = self.config.dict["capture_volume"]["origin_sync_index"]

        # QC needed to get the corner distance accuracy within the GUI
        # Note that the corner distance accuracy calcs need validation...I'm not relying on them now...
        self.quality_controller = QualityController(self.capture_volume, charuco=self.charuco)

    def calibrate_capture_volume(self):
        """
        This is where the camera array 6 DoF is set. Many, many things are happening
        here, but they are all necessary steps of the process so I didn't want to
        try to encapsulate any further
        """

        def worker():
            output_path = Path(self.workspace_guide.extrinsic_dir, "CHARUCO", "xy_CHARUCO.csv")
            if output_path.exists():
                output_path.unlink()  # make sure this doesn't exist to begin with.

            self.load_extrinsic_stream_manager()

            # config settings that help to throttle processing rate to manage resource demands
            include_video = self.config.get_save_tracked_points()
            fps_target = self.config.get_fps_sync_stream_processing()

            self.extrinsic_stream_manager.process_streams(fps_target=fps_target, include_video=include_video)
            logger.info(f"Processing of extrinsic calibration begun...waiting for output to populate: {output_path}")

            logger.info("About to signal that synched frames should be shown")
            self.show_synched_frames.emit()

            while not output_path.exists():
                sleep(0.5)
                # moderate the frequency with which logging statements get made
                if round(time()) % 3 == 0:
                    logger.info(f"Waiting for 2D tracked points to populate at {output_path}")

            # note that this processing will wait until it is complete
            # self.process_extrinsic_streams(fps_target=100)
            logger.info("Processing if extrinsic caliberation streams complete...")

            self.extrinsic_calibration_xy = Path(
                self.workspace, "calibration", "extrinsic", "CHARUCO", "xy_CHARUCO.csv"
            )

            stereocalibrator = StereoCalibrator(self.config.config_toml_path, self.extrinsic_calibration_xy)
            stereocalibrator.stereo_calibrate_all(boards_sampled=10)

            # refreshing camera array from config file
            self.camera_array: CameraArray = CameraArrayInitializer(
                self.config.config_toml_path
            ).get_best_camera_array()

            self.point_estimates: PointEstimates = get_point_estimates(self.camera_array, self.extrinsic_calibration_xy)

            self.capture_volume = CaptureVolume(self.camera_array, self.point_estimates)
            self.capture_volume.optimize()

            self.quality_controller = QualityController(self.capture_volume, self.charuco)

            logger.info(f"Removing the worst fitting {FILTERED_FRACTION*100} percent of points from the model")
            self.quality_controller.filter_point_estimates(FILTERED_FRACTION)
            self.capture_volume.optimize()

            # saves both point estimates and camera array
            self.config.save_capture_volume(self.capture_volume)

        self.calibrate_capture_volume_thread = QThread()
        self.calibrate_capture_volume_thread.run = worker
        self.calibrate_capture_volume_thread.finished.connect(self.capture_volume_calibrated.emit)
        self.calibrate_capture_volume_thread.start()

    def process_recordings(self, recording_path: Path, tracker_enum: TrackerEnum):
        """
        Initiates worker thread to begin post processing.
        TrackerEnum passed in so that access is given to both the tracker and the name
        The name is needed for file/folder naming
        """

        def worker():
            logger.info(f"Beginning to process video files at {recording_path}")
            logger.info(f"Creating post processor for {recording_path}")
            self.post_processor = PostProcessor(self.camera_array, recording_path, tracker_enum)

            # config settings that help to throttle processing rate to manage resource demands
            include_video = self.config.get_save_tracked_points()
            fps_target = self.config.get_fps_sync_stream_processing()

            self.post_processor.create_xy(include_video=include_video, fps_target=fps_target)
            self.post_processor.create_xyz()

        self.process_recordings_thread = QThread()
        self.process_recordings_thread.run = worker
        self.process_recordings_thread.finished.connect(self.post_processing_complete)
        self.process_recordings_thread.start()

    def rotate_capture_volume(self, direction: str):
        transformations = {
            "x+": np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, -1, 0, 0], [0, 0, 0, 1]], dtype=float),
            "x-": np.array([[1, 0, 0, 0], [0, 0, -1, 0], [0, 1, 0, 0], [0, 0, 0, 1]], dtype=float),
            "y+": np.array([[0, 0, -1, 0], [0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 1]], dtype=float),
            "y-": np.array([[0, 0, 1, 0], [0, 1, 0, 0], [-1, 0, 0, 0], [0, 0, 0, 1]], dtype=float),
            "z+": np.array([[0, 1, 0, 0], [-1, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], dtype=float),
            "z-": np.array([[0, -1, 0, 0], [1, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], dtype=float),
        }

        self.capture_volume.shift_origin(transformations[direction])
        self.capture_volume_shifted.emit()

        # don't hold up the rest of the processing just to save the capture volume
        def worker():
            self.config.save_capture_volume(self.capture_volume)

        self.rotate_capture_volume_thread = QThread()
        self.rotate_capture_volume_thread.run = worker
        self.rotate_capture_volume_thread.start()

    def set_capture_volume_origin_to_board(self, origin_index):
        self.capture_volume.set_origin_to_board(origin_index, self.charuco)
        self.capture_volume_shifted.emit()

        def worker():
            self.config.save_capture_volume(self.capture_volume)

        self.set_origin_thread = QThread()
        self.set_origin_thread.run = worker
        self.set_origin_thread.start()

    def autocalibrate(self, port, grid_count, board_threshold):
        def worker():
            self.enable_inputs.emit(port, False)
            self.camera_array.cameras[port].erase_calibration_data()
            self.config.save_camera(self.camera_array.cameras[port])
            self.push_camera_data(port)
            logger.info(f"Initiate autocalibration of grids for port {port}")
            self.intrinsic_stream_manager.autocalibrate(port, grid_count, board_threshold)

            while self.camera_array.cameras[port].matrix is None:
                logger.info(f"Waiting for calibration to complete at port {port}")
                sleep(2)

            self.config.save_camera(self.camera_array.cameras[port])
            self.push_camera_data(port)
            self.intrinsic_stream_manager.stream_jump_to(port, 0)
            self.enable_inputs.emit(port, True)

        self.autocalibrate_threads[port] = QThread()
        self.autocalibrate_threads[port].run = worker
        self.autocalibrate_threads[port].start()



import math

import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl as gl

import caliscope.logger
from caliscope.cameras.camera_array import CameraData

logger = caliscope.logger.get(__name__)


class CameraMesh:
    """Build a camera mesh object that is looking up from the origin"""

    def __init__(self, res, cam_matrix, scale_factor=5000):
        # self.scene = scene

        self.width = res[0] / scale_factor
        self.height = res[1] / scale_factor
        self.fx = cam_matrix[0][0] / scale_factor
        self.fy = cam_matrix[1][1] / scale_factor
        self.cx = cam_matrix[0][2] / scale_factor
        self.cy = cam_matrix[1][2] / scale_factor

        self.f_mean = (self.fx + self.fy) / 2  # mean focal length...height of inverted pyramid

        self.build_verts()
        self.build_faces()
        self.build_colors()

        self.mesh = gl.GLMeshItem(
            vertexes=self.verts,
            faces=self.faces,
            faceColors=self.colors,
            smooth=False,
            drawEdges=True,
            edgeColor=(0, 0, 1, 1),
        )
        self.mesh.setGLOptions("additive")

        logger.debug(self.verts)
        logger.debug(self.faces)
        logger.debug(self.colors)

    def build_verts(self):
        right_side_border = self.width - self.cx
        left_side_border = -self.cx
        top_side_border = self.height - self.cy
        bottom_side_border = -self.cy
        pyramid_height = self.f_mean

        self.verts = [
            [0, 0, 0],  # 0: focal point at origin
            [right_side_border, top_side_border, pyramid_height],  # 1: top right of image
            [right_side_border, bottom_side_border, pyramid_height],  # 2: bottom right of image
            [left_side_border, bottom_side_border, pyramid_height],  # 3: bottom left of image
            [left_side_border, top_side_border, pyramid_height],
        ]  # 4: top left of image

        self.verts = np.array(self.verts)

    def build_faces(self):
        self.faces = [[0, 1, 2], [0, 2, 3], [0, 3, 4], [0, 4, 1], [1, 2, 3], [3, 4, 1]]

        self.faces = np.array(self.faces)

    def build_colors(self):
        self.colors = [
            [0.5, 1, 0, 0.2],
            [0.5, 1, 0, 0.2],
            [0.5, 1, 0, 0.2],
            [0.5, 1, 0, 0.2],
            [0, 0, 0, 0.9],
            [0, 0, 0, 0.9],
        ]

        self.colors = np.array(self.colors)


def rotation_to_float(rotation_matrix):
    new_matrix = []
    for row in rotation_matrix:
        new_row = [float(x) for x in row]
        new_matrix.append(new_row)

    return np.array(new_matrix, dtype=np.float32)


# helper functions to assist with scene creation
def mesh_from_camera(camera_data: CameraData):
    """ "
    Mesh is placed at origin by default. Note that it appears rotations
    are in the mesh frame of reference and translations are in
    the scene frame of reference. I could be wrong, but that appears
    to be the case.

    """
    mesh = CameraMesh(camera_data.size, camera_data.matrix).mesh

    R = camera_data.rotation
    t = camera_data.translation
    camera_orientation_world = R.T

    # rotate mesh
    euler_angles = rotationMatrixToEulerAngles(camera_orientation_world)
    euler_angles_deg = [x * (180 / math.pi) for x in euler_angles]
    x = euler_angles_deg[0]
    y = euler_angles_deg[1]
    z = euler_angles_deg[2]

    # rotate mesh; z,y,x is apparently the order in which it's done
    # https://gamedev.stackexchange.com/questions/16719/what-is-the-correct-order-to-multiply-scale-rotation-and-translation-matrices-f
    mesh.rotate(z, 0, 0, 1, local=True)
    mesh.rotate(y, 0, 1, 0, local=True)
    mesh.rotate(x, 1, 0, 0, local=True)

    camera_origin_world = -np.dot(R.T, t)
    x, y, z = [p for p in camera_origin_world]
    mesh.translate(x, y, z)

    return mesh


def rotationMatrixToEulerAngles(R):
    sy = math.sqrt(R[0, 0] * R[0, 0] + R[1, 0] * R[1, 0])

    singular = sy < 1e-6

    if not singular:
        x = math.atan2(R[2, 1], R[2, 2])
        y = math.atan2(-R[2, 0], sy)
        z = math.atan2(R[1, 0], R[0, 0])
    else:
        x = math.atan2(-R[1, 2], R[1, 1])
        y = math.atan2(-R[2, 0], sy)
        z = 0

    return np.array([x, y, z])


if __name__ == "__main__":
    from pathlib import Path

    import rtoml

    app = pg.mkQApp("GLMeshItem Example")

    scene = gl.GLViewWidget()
    scene.show()
    scene.setWindowTitle("Camera Calibration")
    scene.setCameraPosition(distance=4)

    # grid = gl.GLGridItem()
    # grid.scale(1,1,1)
    # scene.addItem(grid)
    axis = gl.GLAxisItem()
    scene.addItem(axis)

    repo = Path(str(Path(__file__)).split("pyxy")[0], "pyxy")
    # config_path = r"config 2.toml"
    config = rtoml.load(Path(repo, "sessions", "iterative_adjustment", "config.toml"))
    cams = {}
    ports = []

    # build cameras
    for key, params in config.items():
        if "cam" in key:
            res = params["resolution"]
            cam_matrix = params["camera_matrix"]
            port = params["port"]
            ports.append(port)
            for row in cam_matrix:
                for index in range(len(row)):
                    row[index] = float(row[index])

            cam_matrix = np.array(cam_matrix, dtype=np.float32)
            cams[port] = CameraMesh(res, cam_matrix)

            print(key)
            print(params)

    # need to track frame of reference for each camera posiiton
    # must be able ot iterate over each frame of reference
    # place cameras

    origin_port = 0
    cams[origin_port].mesh.setGLOptions("additive")
    scene.addItem(cams[origin_port].mesh)

    for key, params in config.items():
        if "stereo" in key:
            pair = key.split("_")[1:3]
            pair = [int(p) for p in pair]

            if origin_port in pair:
                for param_key, value in params.items():
                    if "translation" in param_key:
                        translation = [float(x[0]) for x in value]
                        translation_scale = 1

                        # reverse translation if origin is the second item
                        if origin_port == pair[1]:
                            translation = [-i for i in translation]
                            other_port = pair[0]
                        else:
                            other_port = pair[1]

                        x, y, z = [t / translation_scale for t in translation]
                        cams[other_port].mesh.translate(x, y, z)
                        logger.info(f"Translation: x: {x}, y: {y}, z: {z}")
                        # cams[other_port].mesh.setGLOptions('additive')

                        scene.addItem(cams[other_port].mesh)
                    if "rotation" in param_key:
                        rotation_count = rotation_to_float(value)  # feeding in 3x3 rotation matrix  from config file
                        rotation_count = rotationMatrixToEulerAngles(rotation_count)  # convert to angles
                        if origin_port == pair[1]:
                            rotation_count = -rotation_count
                            other_port = pair[0]
                        else:
                            other_port = pair[1]

                        rot_deg = [x * (180 / math.pi) for x in rotation_count]  # convert to degrees
                        print(f"Rotation (deg) for port {other_port}: {rot_deg}")
                        x = rot_deg[0]
                        y = rot_deg[1]
                        z = rot_deg[2]

                        logger.info(f"ROTATION: x: {x}, y: {y}, z: {z}")
                        cams[other_port].mesh.rotate(x, 1, 0, 0, local=True)
                        cams[other_port].mesh.rotate(y, 0, 1, 0, local=True)
                        cams[other_port].mesh.rotate(z, 0, 0, 1, local=True)
                        # cams[other_port].mesh.setGLOptions('additive')

                        scene.addItem(cams[other_port].mesh)

    pg.exec()



from enum import Enum

from caliscope.trackers.charuco_tracker import CharucoTracker
from caliscope.trackers.hand_tracker import HandTracker
from caliscope.trackers.holistic.holistic_tracker import HolisticTracker
from caliscope.trackers.pose_tracker import PoseTracker
from caliscope.trackers.simple_holistic_tracker import SimpleHolisticTracker

# Temporarily removed face tracker because crashing sample project
# from caliscope.trackers.face_tracker import FaceTracker


class TrackerEnum(Enum):
    HAND = HandTracker
    POSE = PoseTracker
    SIMPLE_HOLISTIC = SimpleHolisticTracker
    HOLISTIC = HolisticTracker
    CHARUCO = CharucoTracker
    # FACE = FaceTracker


if __name__ == "__main__":
    tracker_factories = [enum_member.name for enum_member in TrackerEnum]
    print(tracker_factories)



from pathlib import Path

import pandas as pd
from PySide6.QtCore import QUrl, Signal
from PySide6.QtGui import QDesktopServices
from PySide6.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

import caliscope.logger
from caliscope.configurator import Configurator
from caliscope.controller import Controller
from caliscope.gui.vizualize.playback_triangulation_widget import (
    PlaybackTriangulationWidget,
)
from caliscope.post_processing.blender_tools import generate_metarig_config
from caliscope.trackers.tracker_enum import TrackerEnum

logger = caliscope.logger.get(__name__)


class PostProcessingWidget(QWidget):
    processing_complete = Signal()

    def __init__(self, controller: Controller):
        super(PostProcessingWidget, self).__init__()
        self.controller = controller
        self.config = self.controller.config

        self.sync_index_cursors = {}  # track where the slider is for each playback...
        self.recording_folders = QListWidget()
        self.update_recording_folders()

        self.vis_widget = PlaybackTriangulationWidget(self.controller.camera_array)

        self.tracker_combo = QComboBox()
        self.vizualizer_title = QLabel()

        # Add items to the combo box using the name attribute of the TrackerEnum
        for tracker in TrackerEnum:
            if tracker.name != "CHARUCO":
                self.tracker_combo.addItem(tracker.name, tracker)

        self.open_folder_btn = QPushButton("&Open Folder")
        self.process_current_btn = QPushButton("&Process")
        self.generate_metarig_config_btn = QPushButton("Generate Metarig Config")
        self.refresh_visualizer()  # must happen before placement to create vis_widget and vizualizer_title
        self.place_widgets()
        self.connect_widgets()

    def set_current_xyz(self):
        if self.xyz_processed_path.exists():
            self.vis_widget.update_motion_trial(self.xyz_processed_path)
        else:
            logger.info(f"No points displayed; Nothing stored in {self.xyz_processed_path}")
            # self.xyz = None

            # check if there aren't any points to track and warn about that
            if self.xy_base_path.exists():
                xy = pd.read_csv(self.xy_base_path)
                if xy.shape[0] == 0:
                    logger.info("No points tracked")
                    QMessageBox.warning(
                        self,
                        "Warning",
                        f"The {self.active_tracker_enum.name} tracker did not identify any points to track in recordings stored in:\n{self.active_recording_path}.",  # noqa 501
                    )  # show a warning dialog

    def update_recording_folders(self):
        # this check here is an artifact of the way that the main widget handles refresh
        self.recording_folders.clear()
        # create list of recording directories
        dir_list = self.controller.workspace_guide.valid_recording_dirs()

        # add each folder to the QListWidget
        for folder in dir_list:
            self.recording_folders.addItem(folder)

        if len(dir_list) > 0:
            self.recording_folders.setCurrentRow(0)

    @property
    def processed_subfolder(self):
        subfolder = Path(
            self.controller.workspace_guide.recording_dir,
            self.recording_folders.currentItem().text(),
            self.tracker_combo.currentData().name,
        )
        return subfolder

    @property
    def xyz_processed_path(self):
        file_name = f"xyz_{self.tracker_combo.currentData().name}.csv"
        result = Path(self.processed_subfolder, file_name)
        return result

    @property
    def archived_config_path(self):
        return Path(self.processed_subfolder, "config.toml")

    @property
    def xy_base_path(self):
        file_name = f"xy_{self.tracker_combo.currentData().name}.csv"
        result = Path(self.processed_subfolder, file_name)
        return result

    @property
    def active_tracker_enum(self):
        return self.tracker_combo.currentData()

    @property
    def metarig_config_path(self):
        file_name = f"metarig_config_{self.tracker_combo.currentData().name}.json"
        result = Path(self.processed_subfolder, file_name)
        return result

    @property
    def active_folder(self):
        if self.recording_folders.count() == 0:
            active_folder = None
        elif self.recording_folders.currentItem() is None:
            self.recording_folders.setCurrentRow(0)
            active_folder: str = self.recording_folders.currentItem().text()
        else:
            active_folder: str = self.recording_folders.currentItem().text()

        return active_folder

    @property
    def active_recording_path(self) -> Path:
        p = Path(self.controller.workspace_guide.recording_dir, self.active_folder)
        logger.info(f"Active recording path is {p}")
        return p

    @property
    def viz_title_html(self):
        if self.xyz_processed_path.exists():
            suffix = "(x,y,z) estimates"
        else:
            suffix = "(no processed data)"

        title = f"<div align='center'><b>{self.tracker_combo.currentData().name.title()} Tracker: {self.active_folder} {suffix} </b></div>"  # noqa E501

        return title

    def place_widgets(self):
        self.setLayout(QHBoxLayout())
        self.left_vbox = QVBoxLayout()
        self.right_vbox = QVBoxLayout()
        self.button_hbox = QHBoxLayout()

        self.layout().addLayout(self.left_vbox)

        self.left_vbox.addWidget(self.recording_folders)
        self.left_vbox.addWidget(self.open_folder_btn)
        self.left_vbox.addWidget(self.tracker_combo)
        self.button_hbox.addWidget(self.process_current_btn)
        self.button_hbox.addWidget(self.generate_metarig_config_btn)
        self.left_vbox.addLayout(self.button_hbox)

        self.layout().addLayout(self.right_vbox, stretch=2)
        self.right_vbox.addWidget(self.vizualizer_title)
        self.right_vbox.addWidget(self.vis_widget, stretch=2)

    def connect_widgets(self):
        self.recording_folders.currentItemChanged.connect(self.refresh_visualizer)
        self.tracker_combo.currentIndexChanged.connect(self.refresh_visualizer)
        self.vis_widget.slider.valueChanged.connect(self.store_sync_index_cursor)
        self.process_current_btn.clicked.connect(self.process_current)
        self.open_folder_btn.clicked.connect(self.open_folder)
        self.generate_metarig_config_btn.clicked.connect(self.create_metarig_config)

        self.controller.post_processing_complete.connect(self.enable_all_inputs)
        self.controller.post_processing_complete.connect(self.refresh_visualizer)

    def store_sync_index_cursor(self, cursor_value):
        if self.xyz_processed_path.exists():
            self.sync_index_cursors[self.xyz_processed_path] = cursor_value

    def open_folder(self):
        """Opens the currently active folder in a system file browser"""
        if self.active_folder is not None:
            folder_path = Path(self.controller.workspace_guide.recording_dir, self.active_folder)
            url = QUrl.fromLocalFile(str(folder_path))
            QDesktopServices.openUrl(url)
        else:
            logger.warning("No folder selected")

    def process_current(self):
        """

        This needs to get pushed into the controller layer
        """
        recording_path = Path(self.controller.workspace_guide.recording_dir, self.active_folder)
        logger.info(f"Beginning processing of recordings at {recording_path}")
        tracker_enum = self.tracker_combo.currentData()
        logger.info(f"(x,y) tracking will be applied using {tracker_enum.name}")
        recording_config_toml = Path(recording_path, "config.toml")
        logger.info(f"Camera data based on config file saved to {recording_config_toml}")

        self.controller.process_recordings(recording_path, tracker_enum)

    def active_tracker_config_path(self):
        """
        This will exist of the tracker has been processed. It should have a camera array within it
        that can be displayed to the user and will align
        """
        return Path(self.active_recording_path, self.active_tracker_enum.name, "config.toml")

    def refresh_visualizer(self):
        logger.info("Refreshing vizualizer within post_processing widget")

        if (
            self.archived_config_path.exists()
        ):  # processing has been done and their is a camera array that can be loaded
            stored_config = Configurator(self.archived_config_path.parent)
            presented_camera_array = stored_config.get_camera_array()
        else:
            presented_camera_array = self.controller.camera_array

        self.vis_widget.update_camera_array(presented_camera_array)

        self.set_current_xyz()
        self.vizualizer_title.setText(self.viz_title_html)
        self.update_enabled_disabled()
        self.update_slider_position()

    def disable_all_inputs(self):
        """used to toggle off all inputs will processing is going on"""
        self.recording_folders.setEnabled(False)
        self.tracker_combo.setEnabled(False)
        # self.export_btn.setEnabled(False)
        self.process_current_btn.setEnabled(False)
        self.vis_widget.slider.setEnabled(False)

    def enable_all_inputs(self):
        """
        after processing completes, swithes everything on again,
        but fine tuning of enable/disable will happen with self.update_enabled_disabled
        """
        self.recording_folders.setEnabled(True)
        self.tracker_combo.setEnabled(True)
        self.process_current_btn.setEnabled(True)
        self.vis_widget.slider.setEnabled(True)

    def update_enabled_disabled(self):
        # set availability of metarig generation
        logger.info("Checking if metarig config can be created...")
        tracker = self.tracker_combo.currentData().value()
        logger.info(tracker)
        if tracker.metarig_mapped and self.xyz_processed_path.exists() and not self.metarig_config_path.exists():
            self.generate_metarig_config_btn.setEnabled(True)
            self.generate_metarig_config_btn.setToolTip("Creation of metarig configuration file is now available")
        else:
            self.generate_metarig_config_btn.setEnabled(False)

        if not tracker.metarig_mapped:
            self.generate_metarig_config_btn.setToolTip("Tracker is not set up to scale to a metarig")
        elif self.metarig_config_path.exists():
            self.generate_metarig_config_btn.setToolTip(
                "The Metarig configuration json file has already been created. Check the tracker subfolder in the recording directory."  # noqa E501
            )
        elif not self.xyz_processed_path.exists():
            self.generate_metarig_config_btn.setToolTip(
                "Must process recording to create xyz estimates for metarig configuration"
            )
        else:
            self.generate_metarig_config_btn.setToolTip(
                "Click to create a file in the tracker subfolder that can be used to scale a Blender metarig"
            )
        # set availability of Proecssing and slider
        if self.xyz_processed_path.exists():
            self.process_current_btn.setEnabled(False)
            self.vis_widget.slider.setEnabled(True)
        elif self.xy_base_path.exists() and not self.xyz_processed_path.exists():
            # nothing available to triangulate
            self.process_current_btn.setEnabled(False)
            self.vis_widget.slider.setEnabled(False)

        else:
            self.process_current_btn.setEnabled(True)
            self.vis_widget.slider.setEnabled(False)

    def update_slider_position(self):
        # update slider value to stored value if it exists
        if self.xyz_processed_path in self.sync_index_cursors.keys():
            active_sync_index = self.sync_index_cursors[self.xyz_processed_path]
            self.vis_widget.slider.setValue(active_sync_index)
            self.vis_widget.visualizer.display_points(active_sync_index)
        else:
            pass

    def create_metarig_config(self):
        logger.info(f"Beginning metarig_config creation in {self.processed_subfolder}")
        tracker_enum = self.tracker_combo.currentData()
        xyz_csv_path = Path(self.processed_subfolder, f"xyz_{tracker_enum.name}_labelled.csv")
        generate_metarig_config(tracker_enum, xyz_csv_path)
        self.update_enabled_disabled()



from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
import pyqtgraph as pg
from pyqtgraph.opengl import GLLinePlotItem

from caliscope.packets import PointPacket, XYZPacket


class Tracker(ABC):
    @property
    def name(self) -> str:
        """
        returns the tracker name
        This name should align with the label used by TrackerEnum
        Used for file naming creation
        """
        pass

    @abstractmethod
    def get_points(self, frame: np.ndarray, port: int, rotation_count: int) -> PointPacket:
        """
        frame: np.ndarray from reading an OpenCV capture object

        port: integer value usd to track which camera the frame originates from

        rotation count: used to indicate the orientation of the image (e.g. rotateed 90 degrees left or right)
                        Some tracking algorithms expect images to be "upright", so this can be used to align the image

                        The function `apply_rotation` from `caliscope.trackers.helper` can correctly orient the image
                        The function `unrotate_points` from the same module can convert any tracked points back into
                        the original orientation
        """
        pass

    @abstractmethod
    def get_point_name(self, point_id: int) -> str:
        """
        Maps point_id to a name
        Used for saving out data with sensible headers.
        """
        pass

    @abstractmethod
    def scatter_draw_instructions(self, point_id: int) -> dict:
        """
        Maps point_id to a dictionary of parameters used to draw circles on frames for visual feedback.

        As an example, the dictionary could have the form: {"radius": 5, "color": (220, 0, 0), "thickness": 3}
        The parameters `radius`, `color`, and `thickness` are used downstream in a call to `cv2.circle`
        to place the tracked point on the frame. See `FramePacket` below.
        """
        pass

    def get_connected_points(self) -> set[tuple[int, int]]:
        """
        OPTIONAL METHOD
        used for 2d drawing purposes elsewhere. Specify which
        points (if any) should have a line connecting them
        {(point_id_A, point_id_B),etc...}

        currently only implemented for charuco...
        """
        pass

    @property
    def metarig_mapped(self):
        """
        OPTIONAL PROPERTY

        Defaults to false and can be overriden to True
        Used to ensure that metarig_config creation is not presented as
        an option in GUI
        """
        return False

    @property
    def metarig_symmetrical_measures(self):
        """
        OPTIONAL PROPERTY

        a dictionary of key: value in the form Measure_Name:[pointA, pointB]
        when processed, the mean distances (excluding outliers) of both
        left_pointA,left_pointB and right_pointA, right_pointB will be calculated.
        The mean of the two sides will be taken
        """
        raise NotImplementedError(f"Tracker {self.name} has not provided its measures for configuring a metarig")

    @property
    def metarig_bilateral_measures(self):
        """
        OPTIONAL PROPERTY

        a dictionary of key: value in the form Measure_Name:[side_pointA, side_pointB]
        when processed, mean distance (excluding outliers) between the two points
        will be calculated and stored as the measure
        """
        raise NotImplementedError(f"Tracker {self.name} has not provided its measures for configuring a metarig")


@dataclass(slots=True, frozen=True)
class Segment:
    name: str
    color: str  # one of: r, g, b, c, m, y, k, w
    point_A: str  # name of landmark
    point_B: str  # name of landmark
    width: float = 1  # note that this does not scale with zoom level... should probably just stick with 1


@dataclass(slots=False, frozen=False)
class WireFrameView:
    segments: [Segment]
    point_names: dict[str:int]  # map landmark name to landmark id

    def __post_init__(self):
        self.line_plots = {}
        for segment in self.segments:
            self.line_plots[segment.name] = GLLinePlotItem(
                color=pg.mkColor(segment.color), width=segment.width, mode="lines"
            )

        self.point_ids = {value: key for key, value in self.point_names.items()}

    def set_points(self, xyz_packet: XYZPacket):
        for segment in self.segments:
            A_id = self.point_ids[segment.point_A]
            B_id = self.point_ids[segment.point_B]
            self.line_plots[segment.name].setData(pos=xyz_packet.get_segment_ends(A_id, B_id))



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDoubleSpinBox


def calculate_spinbox_width(spin_box, min_width=None, padding=None):
    """
    Calculate appropriate width for a spin box based on its content.

    Args:
        spin_box: QSpinBox or QDoubleSpinBox instance
        min_width: Optional minimum width to enforce
        padding: Optional padding to add to calculated width

    Returns:
        Calculated minimum width in pixels
    """
    fm = spin_box.fontMetrics()

    # Get the maximum possible string length
    if isinstance(spin_box, QDoubleSpinBox):
        # For double, consider decimal places
        decimals = spin_box.decimals()
        # Format with maximum digits: -1234.56
        max_str = f"{spin_box.minimum():.{decimals}f}"
        min_str = f"{spin_box.maximum():.{decimals}f}"
    else:
        max_str = str(spin_box.minimum())
        min_str = str(spin_box.maximum())

    # Get width of the longest possible string
    max_width = max(fm.horizontalAdvance(max_str),
                   fm.horizontalAdvance(min_str))

    # Add space for spin arrows and frame
    if padding is None:
        padding = fm.height() + 20  # Default padding based on font height

    width = max_width + padding

    # Enforce minimum width if specified
    if min_width is not None:
        width = max(width, min_width)

    return width

def setup_spinbox_sizing(spin_box, centered=True, min_value=None, max_value=None, min_width=None, padding=None):
    """
    Configure sizing and alignment for a spin box.

    Args:
        spin_box: QSpinBox or QDoubleSpinBox instance
        centered: Whether to center-align the text
        min_width: Optional minimum width to enforce
        padding: Optional padding to add to calculated width
    """
    if centered:
        spin_box.setAlignment(Qt.AlignmentFlag.AlignCenter)

    if min_value is not None:
        spin_box.setMinimum(min_value)
    if max_value is not None:
        spin_box.setMaximum(max_value)

    width = calculate_spinbox_width(spin_box, min_width, padding)
    spin_box.setMinimumWidth(width)

    return width



from time import time

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QApplication,
    QPushButton,
    QTextBrowser,
    QVBoxLayout,
    QWidget,
)

from caliscope.logger import XStream, get

logger = get(__name__)


class LogWidget(QWidget):
    def __init__(self, message: str = None):
        super(LogWidget, self).__init__()
        self.setWindowTitle(message)
        self._console = LogMessageViewer(self)

        self.setWindowFlags(Qt.WindowType.WindowTitleHint)

        layout = QVBoxLayout()

        layout.addWidget(self._console)

        ## Verify widget working with a button to send a message
        if __name__ == "__main__":
            self._button = QPushButton(self)
            self._button.setText("Test Me")
            layout.addWidget(self._button)
            self._button.clicked.connect(test)

        self.setLayout(layout)
        XStream.stdout().messageWritten.connect(self._console.appendLogMessage)
        XStream.stderr().messageWritten.connect(self._console.appendLogMessage)


def test():
    logger.info(f"This is a test; It is {time()}")


class LogMessageViewer(QTextBrowser):
    def __init__(self, parent=None):
        super(LogMessageViewer, self).__init__(parent)
        self.setReadOnly(True)
        # self.setLineWrapMode(QtGui.QTextEdit.NoWrap)
        self.setEnabled(True)
        self.verticalScrollBar().setVisible(True)

    @Slot(str)
    def appendLogMessage(self, msg):
        horScrollBar = self.horizontalScrollBar()
        verScrollBar = self.verticalScrollBar()
        # scrollIsAtEnd = verScrollBar.maximum() - verScrollBar.value() <= 10

        verScrollBar.setValue(verScrollBar.maximum())  # Scrolls to the bottom
        horScrollBar.setValue(0)  # scroll to the left
        self.insertPlainText(msg)

        # if scrollIsAtEnd:
        #     verScrollBar.setValue(verScrollBar.maximum())  # Scrolls to the bottom
        #     self.insertPlainText(msg)
        #     horScrollBar.setValue(0)  # scroll to the left


if __name__ == "__main__":
    app = QApplication([])
    dlg = LogWidget("This is only a test")
    dlg.show()

    app.exec()



from pathlib import Path

import caliscope.logger
from caliscope.configurator import Configurator

logger = caliscope.logger.get(__name__)


class WorkspaceGuide:
    def __init__(self, workspace_dir, camera_count) -> None:
        self.workspace_dir = workspace_dir
        self.camera_count = camera_count
        self.intrinsic_dir = Path(workspace_dir, "calibration", "intrinsic")
        self.extrinsic_dir = Path(workspace_dir, "calibration", "extrinsic")
        self.recording_dir = Path(workspace_dir, "recordings")

    def get_ports_in_dir(self, directory: Path) -> list:
        """
        Returns a list of port indices that are currently exist in calibration/intrinsic
        in the correct file format (i.e. 'port_#.mp4')

        """
        all_ports = []
        for file in directory.iterdir():
            if file.stem[0:5] == "port_" and file.suffix == ".mp4":
                port = file.stem.split("_")[1]
                all_ports.append(int(port))
        return all_ports

    def all_instrinsic_mp4s_available(self):
        return self.missing_files_in_dir(self.intrinsic_dir) == "NONE"

    def all_extrinsic_mp4s_available(self):
        return self.missing_files_in_dir(self.extrinsic_dir) == "NONE"

    def missing_files_in_dir(self, directory: Path):
        files = []
        target_ports = [i for i in range(1, self.camera_count + 1)]
        current_ports = self.get_ports_in_dir(directory)

        missing_ports = [port for port in target_ports if port not in current_ports]
        for port in missing_ports:
            files.append(f"port_{port}.mp4")

        missing_files = ",".join(files)
        if len(missing_files) == 0:
            missing_files = "NONE"
        return missing_files

    def uncalibrated_cameras(self):
        uncalibrated = []
        for cam in self.camera_array.cameras.values():
            if cam.distortions is None and cam.matrix is None and cam.error is None:
                uncalibrated.append(str(cam.port))

        uncalibrated = ",".join(uncalibrated)
        if len(uncalibrated) == 0:
            uncalibrated = "NONE"
        return uncalibrated

    def intrinsic_calibration_status(self):
        if self.camera_array.all_intrinsics_calibrated() and self.all_instrinsic_mp4s_available():
            return "COMPLETE"
        else:
            return "INCOMPLETE"

    def extrinsic_calibration_status(self):
        if self.camera_array.all_extrinsics_calibrated() and self.all_extrinsic_mp4s_available():
            return "COMPLETE"
        else:
            return "INCOMPLETE"

    def valid_recording_dirs(self):
        dir_list = []
        for p in self.recording_dir.iterdir():
            if p.is_dir():
                if self.missing_files_in_dir(p) == "NONE":
                    dir_list.append(p.stem)

        return dir_list

    def valid_recording_dir_text(self) -> str:
        recording_dir_text = ",".join(self.valid_recording_dirs())

        if len(recording_dir_text) == 0:
            recording_dir_text = "NONE"
        return recording_dir_text

    def get_html_summary(self) -> str:
        """
        Provide granular summary of where the workspace is in the calibration process
        Note that the currently configured camera array is reloaded each time this
        is called to determine the state of the data that is currently saved out.
        """
        config = Configurator(self.workspace_dir)
        self.camera_array = config.get_camera_array()
        self.camera_count = config.get_camera_count()

        html = f"""
            <html>
                <head>
                    <style>
                        p {{
                            text-indent: 30px;
                        }}
                    </style>
                </head>
                <body>
                    <h4>Summary</h4>
                    <p>    Directory: {str(self.workspace_dir)}</p>
                    <p>    Camera Count: {self.camera_count}</p>
                    <h4>Intrinsic Calibration: {self.intrinsic_calibration_status()}</h4>
                    <p>    subdirectory: {str(self.intrinsic_dir)}</p>
                    <p>    missing files:{self.missing_files_in_dir(self.intrinsic_dir)}</p>
                    <p>    cameras needing calibration: {self.uncalibrated_cameras()}</p>
                    <h4>Extrinsic Calibration: {self.extrinsic_calibration_status()}</h4>
                    <p>    subdirectory: {str(self.extrinsic_dir)}</p>
                    <p>    missing files:{self.missing_files_in_dir(self.extrinsic_dir)}</p>
                    <h4>Recordings</h4>
                    <p>    valid directories: {self.valid_recording_dir_text()}</p>
                    <p>
                </body>
            </html>
            """

        return html


if __name__ == "__main__":
    workspace_dir = Path(r"C:\Users\Mac Prible\OneDrive\caliscope\4_cam_prerecorded_practice_working")
    camera_count = 4
    workflow_guide = WorkspaceGuide(workspace_dir, camera_count)

    logger.info(workflow_guide.get_html_summary())



import os
import subprocess
import sys
from enum import Enum
from pathlib import Path

import rtoml
from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import (
    QApplication,
    QDockWidget,
    QFileDialog,
    QMainWindow,
    QMenu,
    QTabWidget,
    QWidget,
)

import caliscope.logger
from caliscope import __log_dir__, __root__, __settings_path__
from caliscope.cameras.camera_array import CameraArray
from caliscope.controller import Controller
from caliscope.gui.camera_management.multiplayback_widget import (
    MultiIntrinsicPlaybackWidget,
)
from caliscope.gui.charuco_widget import CharucoWidget
from caliscope.gui.log_widget import LogWidget
from caliscope.gui.post_processing_widget import PostProcessingWidget
from caliscope.gui.vizualize.calibration.capture_volume_visualizer import CaptureVolumeVisualizer
from caliscope.gui.vizualize.calibration.capture_volume_widget import CaptureVolumeWidget
from caliscope.gui.workspace_widget import WorkspaceSummaryWidget

logger = caliscope.logger.get(__name__)


class TabTypes(Enum):
    Workspace = 1
    Charuco = 2
    Cameras = 3
    CaptureVolume = 4


class MainWindow(QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()

        self.app_settings = rtoml.load(__settings_path__)

        self.setWindowTitle("Caliscope")
        self.setWindowIcon(QIcon(str(Path(__root__, "caliscope/gui/icons/box3d-center.svg"))))
        self.setMinimumSize(500, 500)
        self.central_tab = QWidget(self)
        self.setCentralWidget(self.central_tab)

        self.build_menus()
        self.connect_menu_actions()
        self.build_docked_logger()

    def connect_menu_actions(self):
        self.open_project_action.triggered.connect(self.create_new_project_folder)
        self.exit_pyxy3d_action.triggered.connect(QApplication.instance().quit)
        self.open_log_directory_action.triggered.connect(self.open_log_dir)

    def build_menus(self):
        # File Menu
        self.menu = self.menuBar()

        # CREATE FILE MENU
        self.file_menu = self.menu.addMenu("&File")
        self.open_project_action = QAction("New/Open Project", self)
        self.file_menu.addAction(self.open_project_action)

        ####################  Open Recent  ################################
        self.open_recent_project_submenu = QMenu("Recent Projects...", self)

        # Populate the submenu with recent project paths;
        # reverse so that last one appended is at the top of the list
        for project_path in reversed(self.app_settings["recent_projects"]):
            self.add_to_recent_project(project_path)

        self.file_menu.addMenu(self.open_recent_project_submenu)
        ###################################################################

        self.open_log_directory_action = QAction("Open Log Directory")
        self.file_menu.addAction(self.open_log_directory_action)
        self.exit_pyxy3d_action = QAction("Exit", self)
        self.file_menu.addAction(self.exit_pyxy3d_action)

    def build_central_tabs(self):
        self.central_tab = QTabWidget(self)
        self.setCentralWidget(self.central_tab)

        logger.info("Building workspace summary")
        self.workspace_summary = WorkspaceSummaryWidget(self.controller)
        self.workspace_summary.reload_workspace_btn.clicked.connect(self.reload_workspace)
        self.central_tab.addTab(self.workspace_summary, "Workspace")

        if self.controller.all_extrinsic_mp4s_available() and self.controller.camera_array.all_intrinsics_calibrated():
            self.workspace_summary.calibrate_btn.setEnabled(True)
        else:
            self.workspace_summary.calibrate_btn.setEnabled(False)

        logger.info("Building Charuco widget")
        self.charuco_widget = CharucoWidget(self.controller)
        self.central_tab.addTab(self.charuco_widget, "Charuco")

        logger.info("About to load Camera tab")
        if self.controller.cameras_loaded:
            logger.info("Creating MultiIntrinsic Playback Widget")
            self.intrinsic_cal_widget = MultiIntrinsicPlaybackWidget(self.controller)
            logger.info("MultiIntrinsic Playback Widget created")
        else:
            self.intrinsic_cal_widget = QWidget()

        logger.info("finished loading camera tab")
        self.central_tab.addTab(self.intrinsic_cal_widget, "Cameras")
        self.central_tab.setTabEnabled(self.find_tab_index_by_title("Cameras"), self.controller.cameras_loaded)
        logger.info("Camera tab enabled")

        logger.info("About to load capture volume tab")
        if self.controller.capture_volume_loaded:
            logger.info("Creating capture Volume Widget")
            self.capture_volume_widget = CaptureVolumeWidget(self.controller)
        else:
            logger.info("Creating dummy widget")
            self.capture_volume_widget = QWidget()
        self.central_tab.addTab(self.capture_volume_widget, "Capture Volume")
        self.central_tab.setTabEnabled(
            self.find_tab_index_by_title("Capture Volume"), self.controller.capture_volume_loaded
        )

        logger.info("About to load post-processing tab")
        if self.controller.capture_volume_loaded and self.controller.recordings_available():
            logger.info("Creating post processing widget")
            self.post_processing_widget = PostProcessingWidget(self.controller)
            self.controller.capture_volume_shifted.connect(self.post_processing_widget.refresh_visualizer)
            post_processing_enabled = True
        else:
            logger.info("Creating dummy widget")
            self.post_processing_widget = QWidget()
            post_processing_enabled = False
        self.central_tab.addTab(self.post_processing_widget, "Post Processing")
        self.central_tab.setTabEnabled(self.find_tab_index_by_title("Post Processing"), post_processing_enabled)

    def build_docked_logger(self):
        # create log window which is fixed below main window
        self.docked_logger = QDockWidget("Log", self)
        self.docked_logger.setFeatures(QDockWidget.DockWidgetFeature.DockWidgetMovable)
        self.docked_logger.setAllowedAreas(Qt.DockWidgetArea.BottomDockWidgetArea)
        self.log_widget = LogWidget()
        self.docked_logger.setWidget(self.log_widget)

        self.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, self.docked_logger)

    def launch_workspace(self, path_to_workspace: str):
        logger.info(f"Launching session with config file stored in {path_to_workspace}")
        self.controller = Controller(Path(path_to_workspace))
        self.controller.load_workspace_thread.finished.connect(self.build_central_tabs)
        logger.info("Initiate controller loading")
        self.controller.load_workspace()

        self.open_project_action.setEnabled(False)
        self.open_recent_project_submenu.setEnabled(False)

    def find_tab_index_by_title(self, title):
        # Iterate through tabs to find the index of the tab with the given title
        for index in range(self.central_tab.count()):
            if self.central_tab.tabText(index) == title:
                return index
        return -1  # Return -1 if the tab is not found

    def reload_workspace(self):
        # Clear all existing tabs
        logger.info("Clearing workspace")
        # Iterate backwards through the tabs and remove them
        for index in range(self.central_tab.count() - 1, -1, -1):
            widget_to_remove = self.central_tab.widget(index)
            logger.info(f"Removing tab with index {index}")
            self.central_tab.removeTab(index)
            if widget_to_remove is not None:
                widget_to_remove.deleteLater()

            self.central_tab.clear()

        workspace = self.controller.workspace
        del self.controller
        self.controller = Controller(workspace_dir=workspace)
        self.controller.load_workspace()
        self.controller.load_workspace_thread.finished.connect(self.build_central_tabs)

    def add_to_recent_project(self, project_path: str):
        recent_project_action = QAction(project_path, self)
        recent_project_action.triggered.connect(self.open_recent_project)
        self.open_recent_project_submenu.addAction(recent_project_action)

    def open_recent_project(self):
        action = self.sender()
        project_path = action.text()
        logger.info(f"Opening recent session stored at {project_path}")
        self.launch_workspace(project_path)

    def open_log_dir(self):
        logger.info(f"Opening logging directory within File Explorer...  located at {__log_dir__}")
        if sys.platform == "win32":
            os.startfile(__log_dir__)
        elif sys.platform == "darwin":
            subprocess.run(["open", __log_dir__])
        else:  # Linux and Unix-like systems
            subprocess.run(["xdg-open", __log_dir__])
        pass

    def create_new_project_folder(self):
        default_folder = Path(self.app_settings["last_project_parent"])
        dialog = QFileDialog()
        path_to_folder = dialog.getExistingDirectory(
            parent=None,
            caption="Open Previous or Create New Project Directory",
            dir=str(default_folder),
            options=QFileDialog.Option.ShowDirsOnly,
        )

        if path_to_folder:
            logger.info(("Creating new project in :", path_to_folder))
            self.add_project_to_recent(path_to_folder)
            self.launch_workspace(path_to_folder)

    def add_project_to_recent(self, folder_path):
        if str(folder_path) in self.app_settings["recent_projects"]:
            pass
        else:
            self.app_settings["recent_projects"].append(str(folder_path))
            self.app_settings["last_project_parent"] = str(Path(folder_path).parent)
            self.update_app_settings()
            self.add_to_recent_project(folder_path)

    def update_app_settings(self):
        logger.info(f"Saving out app settings to {__settings_path__}")
        with open(__settings_path__, "w") as f:
            rtoml.dump(self.app_settings, f)


def launch_main():
    # import qdarktheme

    app = QApplication(sys.argv)
    dummy_widget = CaptureVolumeVisualizer(camera_array=CameraArray({}))  #  try to force "blinking to initial main"
    del dummy_widget
    # qdarktheme.setup_theme("auto")
    window = MainWindow()
    window.show()
    app.exec()


if __name__ == "__main__":
    launch_main()
    # pass



from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSlider,
    QVBoxLayout,
    QWidget,
)

import caliscope.logger
from caliscope.controller import Controller
from caliscope.gui.vizualize.calibration.capture_volume_visualizer import (
    CaptureVolumeVisualizer,
)

logger = caliscope.logger.get(__name__)


class CaptureVolumeWidget(QWidget):
    def __init__(self, controller: Controller):
        super(CaptureVolumeWidget, self).__init__()

        self.controller = controller

        if not hasattr(self.controller, "capture_volume"):
            self.controller.load_estimated_capture_volume()

        self.visualizer = CaptureVolumeVisualizer(self.controller.capture_volume)
        # self.visualizer.scene.show()
        self.slider = QSlider(Qt.Orientation.Horizontal)
        self.slider.setMinimum(self.visualizer.min_sync_index)
        self.slider.setMaximum(self.visualizer.max_sync_index)
        self.set_origin_btn = QPushButton("Set Origin")

        self.setMinimumSize(500, 500)

        self.rotate_x_plus_btn = QPushButton("X+")
        self.rotate_x_minus_btn = QPushButton("X-")
        self.rotate_y_plus_btn = QPushButton("Y+")
        self.rotate_y_minus_btn = QPushButton("Y-")
        self.rotate_z_plus_btn = QPushButton("Z+")
        self.rotate_z_minus_btn = QPushButton("Z-")

        self.rmse_summary = QLabel(self.controller.capture_volume.get_rmse_summary())

        self.place_widgets()
        self.connect_widgets()

        self.visualizer.display_points(self.visualizer.min_sync_index)

    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.visualizer.scene, stretch=2)
        self.layout().addWidget(self.slider)

        self.grid = QGridLayout()
        self.grid.addWidget(self.rotate_x_plus_btn, 0, 0)
        self.grid.addWidget(self.rotate_x_minus_btn, 1, 0)
        self.grid.addWidget(self.rotate_y_plus_btn, 0, 1)
        self.grid.addWidget(self.rotate_y_minus_btn, 1, 1)
        self.grid.addWidget(self.rotate_z_plus_btn, 0, 2)
        self.grid.addWidget(self.rotate_z_minus_btn, 1, 2)

        self.world_origin_group = QGroupBox()
        self.world_origin_group.setLayout(QVBoxLayout())
        self.world_origin_group.layout().addWidget(self.set_origin_btn)
        self.world_origin_group.layout().addLayout(self.grid)

        self.calibrate_group = QGroupBox()
        self.calibrate_group.setLayout(QVBoxLayout())
        self.calibrate_group.layout().addWidget(self.rmse_summary)
        # self.calibrate_group.layout().addWidget(self.recalibrate_btn)

        self.hbox = QHBoxLayout()
        self.hbox.addWidget(self.calibrate_group)
        self.hbox.addWidget(self.world_origin_group)
        self.layout().addLayout(self.hbox)

        # self.layout().addWidget(self.recalibrate_btn)

    def connect_widgets(self):
        self.slider.valueChanged.connect(self.visualizer.display_points)
        self.set_origin_btn.clicked.connect(self.set_origin_to_board)
        self.rotate_x_plus_btn.clicked.connect(lambda: self.rotate_capture_volume("x+"))
        self.rotate_x_minus_btn.clicked.connect(lambda: self.rotate_capture_volume("x-"))
        self.rotate_y_plus_btn.clicked.connect(lambda: self.rotate_capture_volume("y+"))
        self.rotate_y_minus_btn.clicked.connect(lambda: self.rotate_capture_volume("y-"))
        self.rotate_z_plus_btn.clicked.connect(lambda: self.rotate_capture_volume("z+"))
        self.rotate_z_minus_btn.clicked.connect(lambda: self.rotate_capture_volume("z-"))

    def set_origin_to_board(self):
        logger.info("Setting origin to board...")
        origin_index = self.slider.value()
        self.controller.set_capture_volume_origin_to_board(origin_index)
        self.visualizer.refresh_scene()

    def rotate_capture_volume(self, direction):
        logger.info(f"Rotating capture volume: {direction}")

        self.controller.rotate_capture_volume(direction)
        self.visualizer.refresh_scene()

    def update_board(self, sync_index):
        logger.info(f"Updating board to sync index {sync_index}")

        self.visualizer.display_points(sync_index)



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QLabel, QProgressBar, QVBoxLayout


class ProgressDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle("Landmark Tracking and Triangulation")

        self.display_text = QLabel()
        self.progress_bar = QProgressBar(self)
        # self.progress_bar.setGeometry(0, 0, 300, 25)
        # self.progress_bar.setMaximum(100)
        self.progress_bar.setAlignment(Qt.AlignmentFlag.AlignCenter)
        # self.progress_bar.setFormat("This is a test")

        layout = QVBoxLayout(self)
        layout.addWidget(self.display_text)
        layout.addWidget(self.progress_bar)

    def update(self, data: dict):
        if "close" in data.keys():
            self.hide()
        else:
            self.display_text.setText(data["stage"])
            self.progress_bar.setValue(data["percent"])



# %%

# NOTE: Conversions are being made here between inches and cm because
# this seems like a reasonable scale for discussing the board, but when
# it is actually created in OpenCV, the board height is expressed
# in meters as a standard convention of science, and to improve
# readability of 3D positional output downstream

from collections import defaultdict
from itertools import combinations

import cv2
from PySide6.QtCore import Qt
from PySide6.QtGui import QImage, QPixmap

import caliscope.logger

logger = caliscope.logger.get(__name__)

INCHES_PER_CM = 0.393701

class Charuco:
    """
    create a charuco board that can be printed out and used for camera
    calibration, and used for drawing a grid during calibration
    """

    def __init__(
        self,
        columns,
        rows,
        board_height,
        board_width,
        dictionary="DICT_4X4_50",
        units="inch",
        aruco_scale=0.75,
        square_size_overide_cm=None,
        inverted=False,
        legacy_pattern=False
    ):  # after printing, measure actual and return to overide
        """
        Create board based on shape and dimensions
        square_size_overide_cm: correct for the actual printed size of the board
        """
        self.columns = columns
        self.rows = rows

        self.board_height = board_height
        self.board_width = board_width
        self.dictionary = dictionary

        self.units = units
        self.aruco_scale = aruco_scale
        # if square length not provided, calculate based on board dimensions
        # to maximize size of squares
        self.square_size_overide_cm = square_size_overide_cm
        self.inverted = inverted
        self.legacy_pattern = legacy_pattern

    @property
    def board_height_cm(self):
        """Internal calculations will always use mm for consistency"""
        if self.units == "inch":
            return self.board_height / INCHES_PER_CM
        else:
            return self.board_height

    @property
    def board_width_cm(self):
        """Internal calculations will always use mm for consistency"""
        if self.units == "inch":
            return self.board_width / INCHES_PER_CM
        else:
            return self.board_width

    def board_height_scaled(self, pixmap_scale):
        if self.board_height_cm > self.board_width_cm:
            scaled_height = int(pixmap_scale)
        else:
            scaled_height = int(pixmap_scale * (self.board_height_cm/self.board_width_cm))
        return scaled_height

    def board_width_scaled(self, pixmap_scale):
        if self.board_height_cm > self.board_width_cm:
            scaled_width = int(pixmap_scale * (self.board_width_cm/self.board_height_cm))
        else:
            scaled_width = int(pixmap_scale)

        return scaled_width

    @property
    def dictionary_object(self):
        # grab the dictionary from the reference info at the foot of the module
        dictionary_integer = ARUCO_DICTIONARIES[self.dictionary]
        return cv2.aruco.getPredefinedDictionary(dictionary_integer)

    @property
    def board(self):
        if self.square_size_overide_cm:
            square_length = self.square_size_overide_cm / 100  # note: in cm within GUI
        else:
            board_height_m = self.board_height_cm / 100
            board_width_m = self.board_width_cm / 100

            square_length = min([board_height_m / self.rows, board_width_m / self.columns])
        logger.info(f"Creating charuco with square length of {round(square_length,4)}")

        aruco_length = square_length * self.aruco_scale
        # create the board
        board = cv2.aruco.CharucoBoard(size=(self.columns, self.rows),
                                      squareLength= square_length,
                                      markerLength= aruco_length,
                                      dictionary= self.dictionary_object,
        )

        logger.info(f"Setting legacy pattern of board to {self.legacy_pattern}")
        board.setLegacyPattern(self.legacy_pattern)
        return board

    def board_img(self, pixmap_scale=1000):
        """
        returns a cv2 image (numpy array) of the board
        smaller scale image by default for display to GUI
        provide larger max_edge_length to get printer-ready png
        """
        img = self.board.generateImage((self.board_width_scaled(pixmap_scale=pixmap_scale),
                                        self.board_height_scaled(pixmap_scale=pixmap_scale)))
        if self.inverted:
            img = ~img

        return img

    def board_pixmap(self, width, height):
        """
        Convert from an opencv image to QPixmap
        this can be used for creating thumbnail images
        """
        rgb_image = cv2.cvtColor(self.board_img(), cv2.COLOR_BGR2RGB)
        h, w, ch = rgb_image.shape
        bytes_per_line = ch * w
        charuco_QImage = QImage(rgb_image.data, w, h, bytes_per_line, QImage.Format.Format_RGB888)
        p = charuco_QImage.scaled(
            width,
            height,
            Qt.AspectRatioMode.KeepAspectRatio,
            Qt.TransformationMode.SmoothTransformation,
        )
        return QPixmap.fromImage(p)

    def save_image(self, path):
        """
        Saving image at 10x higher resolution than used for GUI
        """
        cv2.imwrite(path, self.board_img(pixmap_scale=10000))

    def save_mirror_image(self, path):
        """
        Saving image at 10x higher resolution than used for GUI
        """
        mirror = cv2.flip(self.board_img(pixmap_scale=10000), 1)
        cv2.imwrite(path, mirror)

    def get_connected_points(self):
        """
        For a given board, returns a set of corner id pairs that will connect to form
        a grid pattern. This will provide the "object points" used by the calibration
        functions. It is the ground truth of how the points relate in the world.

        The return value is a *set* not a list
        """
        # create sets of the vertical and horizontal line positions
        corners = self.board.getChessboardCorners()
        corners_x = corners[:, 0]
        corners_y = corners[:, 1]
        x_set = set(corners_x)
        y_set = set(corners_y)

        lines = defaultdict(list)

        # put each point on the same vertical line in a list
        for x_line in x_set:
            for corner, x, y in zip(range(0, len(corners)), corners_x, corners_y):
                if x == x_line:
                    lines[f"x_{x_line}"].append(corner)

        # and the same for each point on the same horizontal line
        for y_line in y_set:
            for corner, x, y in zip(range(0, len(corners)), corners_x, corners_y):
                if y == y_line:
                    lines[f"y_{y_line}"].append(corner)

        # create a set of all sets of corner pairs that should be connected
        connected_corners = set()
        for lines, corner_ids in lines.items():
            for i in combinations(corner_ids, 2):
                connected_corners.add(i)

        return connected_corners

    def get_object_corners(self, corner_ids):
        """
        Given an array of corner IDs, provide an array of their relative
        position in a board frame of reference, originating from a corner position.
        """

        return self.board.chessboardCorners()[corner_ids, :]

    def summary(self):
        text = f"Columns: {self.columns}\n"
        text = text + f"Rows: {self.rows}\n"
        text = text + f"Board Size: {self.board_width} x {self.board_height} {self.units}\n"
        text = text + f"Inverted:  {self.inverted}\n"
        text = text + "\n"
        text = text + f"Square Edge Length: {self.square_size_overide_cm} cm"
        return text


################################## REFERENCE ###################################
ARUCO_DICTIONARIES = {
    "DICT_4X4_50": cv2.aruco.DICT_4X4_50,
    "DICT_4X4_100": cv2.aruco.DICT_4X4_100,
    "DICT_4X4_250": cv2.aruco.DICT_4X4_250,
    "DICT_4X4_1000": cv2.aruco.DICT_4X4_1000,
    "DICT_5X5_50": cv2.aruco.DICT_5X5_50,
    "DICT_5X5_100": cv2.aruco.DICT_5X5_100,
    "DICT_5X5_250": cv2.aruco.DICT_5X5_250,
    "DICT_5X5_1000": cv2.aruco.DICT_5X5_1000,
    "DICT_6X6_50": cv2.aruco.DICT_6X6_50,
    "DICT_6X6_100": cv2.aruco.DICT_6X6_100,
    "DICT_6X6_250": cv2.aruco.DICT_6X6_250,
    "DICT_6X6_1000": cv2.aruco.DICT_6X6_1000,
    "DICT_7X7_50": cv2.aruco.DICT_7X7_50,
    "DICT_7X7_100": cv2.aruco.DICT_7X7_100,
    "DICT_7X7_250": cv2.aruco.DICT_7X7_250,
    "DICT_7X7_1000": cv2.aruco.DICT_7X7_1000,
    "DICT_ARUCO_ORIGINAL": cv2.aruco.DICT_ARUCO_ORIGINAL,
    "DICT_APRILTAG_16h5": cv2.aruco.DICT_APRILTAG_16h5,
    "DICT_APRILTAG_25h9": cv2.aruco.DICT_APRILTAG_25h9,
    "DICT_APRILTAG_36h10": cv2.aruco.DICT_APRILTAG_36h10,
    "DICT_APRILTAG_36h11": cv2.aruco.DICT_APRILTAG_36h11,
}


if __name__ == "__main__":
    charuco = Charuco(4, 5, 4, 8.5, aruco_scale=0.75, units="inch", inverted=True, square_size_overide_cm=5.25)
    charuco.save_image("test_charuco.png")
    width, height = charuco.board_img().shape
    logger.info(f"Board width is {width}\nBoard height is {height}")

    corners = charuco.board.getChessboardCorners()
    logger.info(corners)

    logger.info(f"Charuco dictionary: {charuco.__dict__}")
    # while True:
    #     cv2.imshow("Charuco Board...'q' to quit", charuco.board_img)
    #     #
    #     key = cv2.waitKey(0)
    #     if key == ord("q"):
    #         cv2.destroyAllWindows()
    #         break

# %%



from queue import Queue
from threading import Thread

import cv2
import mediapipe as mp
import numpy as np

import caliscope.logger
from caliscope.packets import PointPacket
from caliscope.tracker import Tracker
from caliscope.trackers.helper import apply_rotation, unrotate_points

logger = caliscope.logger.get(__name__)

MIN_DETECTION_CONFIDENCE = 0.5
MIN_TRACKING_CONFIDENCE = 0.95

# The following are from base Pose and can be ignored in favor of
# better estimated Holistic points
DRAW_IGNORE_LIST = [
    "nose",
    "left_eye_inner",
    "left_eye",
    "left_eye_outer",
    "right_eye_inner",
    "right_eye",
    "right_eye_outer",
    "left_ear",
    "right_ear",
    "mouth_left",
    "mouth_right",
    "left_wrist_pose",
    "right_wrist_pose",
    "left_pinky",
    "right_pinky",
    "left_index",
    "right_index",
    "left_thumb",
    "right_thumb",
]

POINT_NAMES = {
    # 0: "nose",
    # 1: "left_eye_inner",
    # 2: "left_eye",
    # 3: "left_eye_outer",
    # 4: "right_eye_inner",
    # 5: "right_eye",
    # 6: "right_eye_outer",
    # 7: "left_ear",
    # 8: "right_ear",
    # 9: "mouth_left",
    # 10: "mouth_right",
    11: "left_shoulder",
    12: "right_shoulder",
    13: "left_elbow",
    14: "right_elbow",
    # 15: "left_wrist_pose",
    # 16: "right_wrist_pose",
    # 17: "left_pinky",
    # 18: "right_pinky",
    # 19: "left_index",
    # 20: "right_index",
    # 21: "left_thumb",
    # 22: "right_thumb",
    23: "left_hip",
    24: "right_hip",
    25: "left_knee",
    26: "right_knee",
    27: "left_ankle",
    28: "right_ankle",
    29: "left_heel",
    30: "right_heel",
    31: "left_foot_index",
    32: "right_foot_index",
    100: "right_wrist",
    101: "right_thumb_CMC",
    102: "right_thumb_MCP",
    103: "right_thumb_IP",
    104: "right_thumb_tip",
    105: "right_index_finger_MCP",
    106: "right_index_finger_PIP",
    107: "right_index_finger_DIP",
    108: "right_index_finger_tip",
    109: "right_middle_finger_MCP",
    110: "right_middle_finger_PIP",
    111: "right_middle_finger_DIP",
    112: "right_middle_finger_tip",
    113: "right_ring_finger_MCP",
    114: "right_ring_finger_PIP",
    115: "right_ring_finger_DIP",
    116: "right_ring_finger_tip",
    117: "right_pinky_MCP",
    118: "right_pinky_PIP",
    119: "right_pinky_DIP",
    120: "right_pinky_tip",
    200: "left_wrist",
    201: "left_thumb_CMC",
    202: "left_thumb_MCP",
    203: "left_thumb_IP",
    204: "left_thumb_tip",
    205: "left_index_finger_MCP",
    206: "left_index_finger_PIP",
    207: "left_index_finger_DIP",
    208: "left_index_finger_tip",
    209: "left_middle_finger_MCP",
    210: "left_middle_finger_PIP",
    211: "left_middle_finger_DIP",
    212: "left_middle_finger_tip",
    213: "left_ring_finger_MCP",
    214: "left_ring_finger_PIP",
    215: "left_ring_finger_DIP",
    216: "left_ring_finger_tip",
    217: "left_pinky_MCP",
    218: "left_pinky_PIP",
    219: "left_pinky_DIP",
    220: "left_pinky_tip",
    # HOLISTIC FACE keypoints worth keeping...
    # this is for kinematic skull tracking, not animation.
    500: "lip_top_mid",
    504: "nose_tip",
    633: "right_inner_eye",
    699: "chin_tip",
    746: "right_outer_eye",
    862: "left_inner_eye",
    966: "left_outer_eye",
}


METARIG_BILATERAL_MEAUSURES = {
    "Hip_Shoulder_Distance": ["hip", "shoulder"],
    "Shoulder_Inner_Eye_Distance": ["inner_eye", "shoulder"],
    "Palm": ["index_finger_MCP", "pinky_MCP"],
    "Foot": ["heel", "foot_index"],
    "Upper_Arm": ["shoulder", "elbow"],
    "Forearm": ["elbow", "wrist"],
    "Wrist_to_MCP1": ["wrist", "thumb_MCP"],
    "Wrist_to_MCP2": ["wrist", "index_finger_MCP"],
    "Wrist_to_MCP3": ["wrist", "middle_finger_MCP"],
    "Wrist_to_MCP4": ["wrist", "ring_finger_MCP"],
    "Wrist_to_MCP5": ["wrist", "pinky_MCP"],
    "Prox_Phalanx_1": ["thumb_MCP", "thumb_IP"],
    "Prox_Phalanx_2": ["index_finger_MCP", "index_finger_PIP"],
    "Prox_Phalanx_3": ["middle_finger_MCP", "middle_finger_PIP"],
    "Prox_Phalanx_4": ["ring_finger_MCP", "ring_finger_PIP"],
    "Prox_Phalanx_5": ["pinky_MCP", "pinky_PIP"],
    "Mid_Phalanx_2": ["index_finger_PIP", "index_finger_DIP"],
    "Mid_Phalanx_3": ["middle_finger_PIP", "middle_finger_DIP"],
    "Mid_Phalanx_4": ["ring_finger_PIP", "ring_finger_DIP"],
    "Mid_Phalanx_5": ["pinky_PIP", "pinky_DIP"],
    "Dist_Phalanx_1": ["thumb_IP", "thumb_tip"],
    "Dist_Phalanx_2": ["index_finger_DIP", "index_finger_tip"],
    "Dist_Phalanx_3": ["middle_finger_DIP", "middle_finger_tip"],
    "Dist_Phalanx_4": ["ring_finger_DIP", "middle_finger_tip"],
    "Dist_Phalanx_5": ["pinky_DIP", "pinky_tip"],
    "Thigh_Length": ["hip", "knee"],
    "Shin_Length": ["knee", "ankle"],
}


METARIG_SYMMETRICAL_MEASURES = {
    "Shoulder_Width": ["left_shoulder", "right_shoulder"],
    "Hip_Width": ["left_hip", "right_hip"],
    "Inner_Eye_Distance": ["left_inner_eye", "right_inner_eye"],
}


# keep ids in distinct ranges to avoid clashes
POSE_OFFSET = 0
RIGHT_HAND_OFFSET = 100
LEFT_HAND_OFFSET = 200
FACE_OFFSET = 500


class SimpleHolisticTracker(Tracker):
    def __init__(self) -> None:
        # each port gets its own mediapipe context manager
        # use a dictionary of queues for passing
        self.in_queues = {}
        self.out_queues = {}
        self.threads = {}

    @property
    def name(self):
        return "SIMPLE_HOLISTIC"

    @property
    def metarig_mapped(self):
        return True

    @property
    def metarig_symmetrical_measures(self):
        return METARIG_SYMMETRICAL_MEASURES

    @property
    def metarig_bilateral_measures(self):
        return METARIG_BILATERAL_MEAUSURES

    def run_frame_processor(self, port: int, rotation_count: int):
        # Create a MediaPipe pose instance
        with mp.solutions.holistic.Holistic(
            min_detection_confidence=MIN_DETECTION_CONFIDENCE,
            min_tracking_confidence=MIN_TRACKING_CONFIDENCE,
        ) as holistic:
            while True:
                frame = self.in_queues[port].get()
                # apply rotation as needed
                frame = apply_rotation(frame, rotation_count)

                height, width, color = frame.shape
                # Convert the image to RGB format
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                results = holistic.process(frame)

                # initialize variables so none will be created if no points detected
                point_ids = []
                landmark_xy = []

                if results.pose_landmarks:
                    for landmark_id, landmark in enumerate(results.pose_landmarks.landmark):
                        # mediapipe expresses in terms of percent of frame, so must map to pixel position
                        x, y = int(landmark.x * width), int(landmark.y * height)
                        if landmark.x < 0 or landmark.x > 1 or landmark.y < 0 or landmark.y > 1:
                            # ignore
                            pass
                        else:
                            mapped_point_id = landmark_id + POSE_OFFSET
                            # some of the pose values are too noisy to bother with including considering that
                            # holistic face and hand tracking is so good
                            # ignore those points that aren't in the POINT_NAMES list
                            if mapped_point_id in POINT_NAMES:
                                point_ids.append(landmark_id + POSE_OFFSET)
                                landmark_xy.append((x, y))

                if results.right_hand_landmarks:
                    for landmark_id, landmark in enumerate(results.right_hand_landmarks.landmark):
                        # mediapipe expresses in terms of percent of frame, so must map to pixel position
                        x, y = int(landmark.x * width), int(landmark.y * height)
                        if landmark.x < 0 or landmark.x > 1 or landmark.y < 0 or landmark.y > 1:
                            # ignore
                            pass
                        else:
                            point_ids.append(landmark_id + RIGHT_HAND_OFFSET)
                            landmark_xy.append((x, y))

                if results.left_hand_landmarks:
                    for landmark_id, landmark in enumerate(results.left_hand_landmarks.landmark):
                        # mediapipe expresses in terms of percent of frame, so must map to pixel positionND_OFFSET
                        x, y = int(landmark.x * width), int(landmark.y * height)
                        if landmark.x < 0 or landmark.x > 1 or landmark.y < 0 or landmark.y > 1:
                            # ignore
                            pass
                        else:
                            point_ids.append(landmark_id + LEFT_HAND_OFFSET)
                            landmark_xy.append((x, y))

                if results.face_landmarks:
                    for landmark_id, landmark in enumerate(results.face_landmarks.landmark):
                        # mediapipe expresses in terms of percent of frame, so must map to pixel positionFSET
                        x, y = int(landmark.x * width), int(landmark.y * height)
                        if landmark.x < 0 or landmark.x > 1 or landmark.y < 0 or landmark.y > 1:
                            # ignore
                            pass
                        else:
                            face_id = landmark_id + FACE_OFFSET
                            # only track the point if it is in the list of names above
                            # this will significantly reduce the data tracked.
                            if face_id in POINT_NAMES.keys():
                                point_ids.append(landmark_id + FACE_OFFSET)
                                landmark_xy.append((x, y))

                point_ids = np.array(point_ids)
                landmark_xy = np.array(landmark_xy)

                # adjust for previous shift due to camera rotation count
                landmark_xy = unrotate_points(landmark_xy, rotation_count, width, height)
                point_packet = PointPacket(point_ids, landmark_xy)

                self.out_queues[port].put(point_packet)

    def get_points(self, frame: np.ndarray, port: int, rotation_count: int) -> PointPacket:
        """
        This is the primary method exposed to the rest of the code.
        The tracker receives frames and basic camera data from the Stream,
        then it places the frame/camera data on a queue that will hand it
        off to a context manager set up to process that stream of data.
        """

        if port not in self.in_queues.keys():
            self.in_queues[port] = Queue(1)
            self.out_queues[port] = Queue(1)

            self.threads[port] = Thread(
                target=self.run_frame_processor,
                args=(port, rotation_count),
                daemon=True,
            )

            self.threads[port].start()

        self.in_queues[port].put(frame)
        point_packet = self.out_queues[port].get()

        return point_packet

    def get_point_name(self, point_id) -> str:
        # this if/else should be unnecessary now that only select points are being passed on up the chain.
        # if point_id < FACE_OFFSET:
        #     point_name = POINT_NAMES[point_id]
        # else:
        #     point_name = "face_" + str(point_id - FACE_OFFSET)
        return POINT_NAMES[point_id]

    def scatter_draw_instructions(self, point_id: int) -> dict:
        point_name = self.get_point_name(point_id)
        if point_name in DRAW_IGNORE_LIST:
            rules = {"radius": 0, "color": (0, 0, 0), "thickness": 0}
        elif point_name.startswith("left"):
            rules = {"radius": 5, "color": (0, 0, 220), "thickness": 3}
        elif point_name.startswith("right"):
            rules = {"radius": 5, "color": (220, 0, 0), "thickness": 3}
        else:
            rules = {"radius": 3, "color": (0, 220, 220), "thickness": 3}

        return rules



from dataclasses import dataclass

import numpy as np
from scipy.sparse import lil_matrix

import caliscope.logger

logger = caliscope.logger.get(__name__)

CAMERA_PARAM_COUNT = 6


@dataclass
class PointEstimates:
    """
    Initialized from triangulated_points.csv to provide the formatting of data required for bundle adjustment
    "full" is used here because there is currently a method to filter the data based on reprojection error
    Not sure if it will be used going forward, but it remains here if so.
    """

    sync_indices: np.ndarray  # the sync_index from when the image was taken
    camera_indices: np.ndarray  # camera id associated with the img point
    point_id: np.ndarray  # point id (i.e. charuco corner currently)
    img: np.ndarray  # x,y coords of point
    obj_indices: np.ndarray  # mapping of x,y img points to their respective list of estimated x,y,z obj points
    obj: np.ndarray  # x,y,z estimates of object points
    # obj_corner_id: np.ndarray # the charuco corner ID of the xyz object point; is this necessary?

    def __post_init__(self):
        self.sync_indices = self.sync_indices.astype(np.int32)
        self.camera_indices = self.camera_indices.astype(np.int16)
        self.point_id = self.point_id.astype(np.uint16)
        self.img = self.img.astype(np.float64)
        self.obj_indices = self.obj_indices.astype(np.int32)
        self.obj = self.obj.astype(np.float64)

    @property
    def n_cameras(self):
        return np.unique(self.camera_indices).size

    @property
    def n_obj_points(self):
        return self.obj.shape[0]

    @property
    def n_img_points(self):
        return self.img.shape[0]

    def get_sparsity_pattern(self):
        """
        provide the sparsity structure for the Jacobian (elements that are not zero)
        n_points: number of unique 3d points
            these will each have at least one but potentially more associated 2d points

        point_indices: a vector that maps the 2d points to their associated 3d point
        """

        m = self.camera_indices.size * 2
        n = self.n_cameras * CAMERA_PARAM_COUNT + self.n_obj_points * 3
        A = lil_matrix((m, n), dtype=int)

        i = np.arange(self.camera_indices.size)
        for s in range(CAMERA_PARAM_COUNT):
            A[2 * i, self.camera_indices * CAMERA_PARAM_COUNT + s] = 1
            A[2 * i + 1, self.camera_indices * CAMERA_PARAM_COUNT + s] = 1

        for s in range(3):
            A[2 * i, self.n_cameras * CAMERA_PARAM_COUNT + self.obj_indices * 3 + s] = 1
            A[
                2 * i + 1,
                self.n_cameras * CAMERA_PARAM_COUNT + self.obj_indices * 3 + s,
            ] = 1

        return A

    def update_obj_xyz(self, least_sq_result_x):
        """
        Provided with the least_squares estimate of the best fit of model parameters (including camera 6DoF)
        parse out the x,y,z object positions and update self.obj
        """

        xyz = least_sq_result_x[self.n_cameras * CAMERA_PARAM_COUNT :]
        xyz = xyz.reshape(-1, 3)

        self.obj = xyz


def load_point_estimates(config: dict) -> PointEstimates:
    point_estimates_dict = config["point_estimates"]

    for key, value in point_estimates_dict.items():
        point_estimates_dict[key] = np.array(value)

    point_estimates = PointEstimates(**point_estimates_dict)
    return point_estimates



import shutil
from pathlib import Path
from time import sleep

import pandas as pd

import caliscope.logger
from caliscope.cameras.camera_array import CameraArray
from caliscope.export import xyz_to_trc, xyz_to_wide_labelled
from caliscope.post_processing.gap_filling import gap_fill_xy, gap_fill_xyz
from caliscope.post_processing.smoothing import smooth_xyz
from caliscope.synchronized_stream_manager import SynchronizedStreamManager
from caliscope.trackers.tracker_enum import TrackerEnum
from caliscope.triangulate.triangulation import triangulate_xy

logger = caliscope.logger.get(__name__)

# gap filling and filtering is outside the current scope of the project so I'm toggling this off for now
APPLY_EXPERIMENTAL_POST_PROCESSING = False


class PostProcessor:
    """
    The post processer operates independently of the session. It does not need to worry about camera management.
    Provide it with a path to the directory that contains the following:
    - config.toml
    - frame_time.csv
    - .mp4 files

    The post processor will archive the active config.toml file into the subdirectory
    """

    def __init__(self, camera_array: CameraArray, recording_path: Path, tracker_enum: TrackerEnum):
        self.camera_array = camera_array
        self.recording_path = recording_path
        self.tracker_enum = tracker_enum
        self.tracker_name = tracker_enum.name
        self.tracker = tracker_enum.value()

        # save out current camera array to output folder
        tracker_subdirectory = Path(self.recording_path, self.tracker_name)
        tracker_subdirectory.mkdir(exist_ok=True, parents=True)
        shutil.copy(Path(self.recording_path.parent.parent, "config.toml"), Path(tracker_subdirectory, "config.toml"))

        logger.info(f"Creating sync stream manager for videos stored in {self.recording_path}")
        self.sync_stream_manager = SynchronizedStreamManager(
            self.recording_path, self.camera_array.cameras, self.tracker
        )

    def create_xy(self, fps_target=100, include_video=True):
        """
        Reads through all .mp4  files in the recording path and applies the tracker to them
        The xy_TrackerName.csv file is saved out to the same directory by the VideoRecorder

        Note that high fps target and including video will increase processing overhead
        """
        self.sync_stream_manager.process_streams(include_video=include_video, fps_target=fps_target)

        while self.sync_stream_manager.recorder.recording:
            sleep(1)
            percent_complete = int(
                (self.sync_stream_manager.recorder.sync_index / self.sync_stream_manager.mean_frame_count) * 100
            )
            logger.info(f"(Stage 1 of 2): {percent_complete}% of frames processed for (x,y) landmark detection")

    def create_xyz(self, xy_gap_fill=3, xyz_gap_fill=3, cutoff_freq=6, include_trc=True) -> None:
        """
        creates xyz_{tracker name}.csv file within the recording_path directory

        Uses the two functions above, first creating the xy points based on the tracker if they
        don't already exist, the triangulating them. Makes use of an internal method self.triangulate_xy_data

        """

        tracker_output_path = Path(self.recording_path, self.tracker_name)
        xy_csv_path = Path(tracker_output_path, f"xy_{self.tracker_name}.csv")

        # create if it doesn't already exist
        if not xy_csv_path.exists():
            self.create_xy()

        # load in 2d data and triangulate it
        logger.info("Reading in (x,y) data..")
        xy = pd.read_csv(xy_csv_path)
        if xy.shape[0] > 0:
            logger.info("Filling small gaps in (x,y) data")
            xy = gap_fill_xy(xy, max_gap_size=xy_gap_fill)
            logger.info("Beginning data triangulation")
            xyz = triangulate_xy(xy, self.camera_array)
        else:
            logger.warning("No points tracked. Terminating post-processing early.")
            return

        if xyz.shape[0] > 0:
            if APPLY_EXPERIMENTAL_POST_PROCESSING:
                logger.info("Filling small gaps in (x,y,z) data")
                xyz = gap_fill_xyz(xyz, max_gap_size=xyz_gap_fill)
                logger.info("Smoothing (x,y,z) using butterworth filter with cutoff frequency of 6hz")
                xyz = smooth_xyz(xyz, order=2, fps=self.sync_stream_manager.mean_fps, cutoff=cutoff_freq)

            logger.info("Saving (x,y,z) to csv file")
            xyz_csv_path = Path(tracker_output_path, f"xyz_{self.tracker_name}.csv")
            xyz.to_csv(xyz_csv_path)
            xyz_wide_csv_path = Path(tracker_output_path, f"xyz_{self.tracker_name}_labelled.csv")
            xyz_labelled = xyz_to_wide_labelled(xyz, self.tracker_enum.value())
            xyz_labelled.to_csv(xyz_wide_csv_path)

        else:
            logger.warning("No points triangulated. Terminating post-processing early.")
            return

        # only include trc if wanted and only if there is actually good data to export
        if include_trc and xyz.shape[0] > 0:
            trc_path = Path(tracker_output_path, f"xyz_{self.tracker_name}.trc")
            time_history_path = Path(tracker_output_path, "frame_time_history.csv")
            xyz_to_trc(
                xyz,
                tracker=self.tracker_enum.value(),
                time_history_path=time_history_path,
                target_path=trc_path,
            )


if __name__ == "__main__":
    from caliscope.controller import Controller

    workspace_dir = Path(r"C:\Users\Mac Prible\OneDrive - The University of Texas at Austin\research\caliscope\demo")
    controller = Controller(workspace_dir)
    controller.load_camera_array()

    camera_aray = controller.camera_array
    recording_dir = Path(workspace_dir, "recordings", "STS")

    post_processor = PostProcessor(
        camera_array=camera_aray,
        recording_path=recording_dir,
        tracker_enum=TrackerEnum.HAND,
    )

    post_processor.create_xyz()



# %%

from pathlib import Path

import numpy as np
import pandas as pd
import seaborn as sns

from caliscope import __root__

session_directory = Path(__root__, "tests", "demo")


data_2d = pd.read_csv(Path(session_directory, "data_2d.csv"))
distance_error = pd.read_csv(Path(session_directory, "distance_error.csv"))


# %%
facet_reproj_error = sns.FacetGrid(data_2d, col="camera", col_wrap=2)
facet_reproj_error.map_dataframe(sns.boxplot, x="stage", y="reproj_error", showfliers=False)
# %%
facet_distance_error = sns.FacetGrid(distance_error, row="board_distance")
facet_distance_error.map_dataframe(sns.boxplot, x="stage", y="Distance_Error_mm", showfliers=False)
# sns.boxplot(data = distance_error, x = "stage", y = "Distance_Error_mm_abs", showfliers=False)
# %%
facet_distance_error = sns.FacetGrid(distance_error, row="board_distance")
facet_distance_error.map_dataframe(sns.boxplot, x="stage", y="percent_match", showfliers=False)

# %%

by_board_distance = (
    distance_error.filter(["Distance_Error_mm", "Distance_Error_mm_abs", "world_distance", "board_distance"])
    .groupby("board_distance")
    .mean()
)

# %%

rmse = data_2d.filter(["reproj_error_sq"]).mean()

rmse_by_camera = (
    data_2d.filter(["camera", "reproj_error_sq"])
    .groupby("camera")
    .mean(["reproj_error_sq"])
    .rename(columns={"reproj_error_sq": "mean_sq_error"})
)
rmse_by_camera["rmse"] = np.sqrt(rmse_by_camera["mean_sq_error"])

# %%

sns.boxplot(data=distance_error, x="board_distance", y="Distance_Error_mm_abs", showfliers=False)
# %%



"""Top-level package for basic_template_repo."""

import os
import platform
from pathlib import Path

import rtoml

__package_name__ = "caliscope"
__version__ = "v0.2.4"

__author__ = """Mac Prible"""
__email__ = "prible@gmail.com"
__repo_owner_github_user_name__ = "mprib"
__repo_url__ = f"https://github.com/{__repo_owner_github_user_name__}/{__package_name__}/"
__repo_issues_url__ = f"{__repo_url__}issues"


# Determine platform-specific application data directory
if platform.system() == "Windows":
    print("Windows platform identified")
    app_data_dir = os.getenv("LOCALAPPDATA")
else:  # macOS, Linux, and other UNIX variants
    print(f"Non-windows platform identified: {platform.system()}")
    app_data_dir = os.path.join(os.path.expanduser("~"), ".local", "share")

__app_dir__ = Path(app_data_dir, __package_name__)
__app_dir__.mkdir(exist_ok=True, parents=True)

# Create a toml file for user settings in app data directory and default the project folder to USER
__settings_path__ = Path(__app_dir__, "settings.toml")

# Get user home directory in a cross-platform way
__user_dir__ = Path(os.path.expanduser("~"))

if __settings_path__.exists():
    USER_SETTINGS = rtoml.load(__settings_path__)
else:
    # default to storing projects in user/__package_name__
    USER_SETTINGS = {
        "recent_projects": [],
        "last_project_parent": str(
            __user_dir__
        ),  # default initially to home...this will be where the 'New' folder dialog starts
    }

    with open(__settings_path__, "a") as f:
        rtoml.dump(USER_SETTINGS, f)


__log_dir__ = Path(__app_dir__, "logs")
__log_dir__.mkdir(exist_ok=True, parents=True)


# a helpful reference
__root__ = Path(__file__).parent.parent

print(f"This is printing from: {__file__}")
print(f"Source code for this package is available at: {__repo_url__}")
print(f"Log file associated with {__package_name__} is stored in {__log_dir__}")


if __name__ == "__main__":
    import os

    path = r"C:\Users\Mac Prible\AppData\Local\caliscope"
    test_file = os.path.join(path, "test.txt")
    with open(test_file, "w") as f:
        f.write("Test")
    print(f"Test file created at: {test_file}")

    import os

    parent = r"C:\Users\Mac Prible\AppData\Local"
    print(os.path.exists(parent))
    print(os.listdir(parent))



from itertools import combinations

import cv2
import numpy as np

import caliscope.logger
from caliscope.cameras.camera_array import CameraArray, CameraData
from caliscope.triangulate.stereo_points_builder import StereoPointsPacket, SynchedStereoPointsPacket
from caliscope.triangulate.triangulation import undistort

logger = caliscope.logger.get(__name__)


class ArrayStereoTriangulator:
    def __init__(self, camera_array: CameraArray):
        self.camera_array = camera_array

        # pull ports list from camera_array.port_index
        # to ensure only non-ignored cameras are processed
        self.ports = list(camera_array.port_index.keys())
        self.pairs = [(i, j) for i, j in combinations(self.ports, 2) if i < j]

        # create the triangulators for each pair of cameras
        self.triangulators = {}
        for pair in self.pairs:
            port_A = pair[0]
            port_B = pair[1]

            camera_A: CameraData = self.camera_array.cameras[port_A]
            camera_B: CameraData = self.camera_array.cameras[port_B]

            self.triangulators[pair] = StereoPairTriangulator(camera_A, camera_B)

    def triangulate_synched_points(self, synced_paired_points: SynchedStereoPointsPacket):
        for pair, paired_point_packet in synced_paired_points.stereo_points_packets.items():
            if paired_point_packet is not None:
                self.triangulators[pair].add_3D_points(paired_point_packet)


class StereoPairTriangulator:
    def __init__(self, camera_A: CameraData, camera_B: CameraData):
        self.camera_A = camera_A
        self.camera_B = camera_B
        self.portA = camera_A.port
        self.portB = camera_B.port
        self.pair = (self.portA, self.portB)

        self.build_projection_matrices()

    def build_projection_matrices(self):
        # attempting to create something that integrates with the new set_cameras_refactor
        rot_trans_A = np.column_stack([self.camera_A.rotation, self.camera_A.translation])
        mtx_A = self.camera_A.matrix
        self.proj_A = mtx_A @ rot_trans_A  # projection matrix for CamA

        rot_trans_B = np.column_stack([self.camera_B.rotation, self.camera_B.translation])
        mtx_B = self.camera_B.matrix
        self.proj_B = mtx_B @ rot_trans_B  # projection matrix for CamB

    def add_3D_points(self, paired_points: StereoPointsPacket):
        if len(paired_points.common_ids) > 0:
            xy_A = paired_points.img_loc_A
            xy_B = paired_points.img_loc_B

        if xy_A.shape[0] > 0:
            logger.info(f"Triangulating points in common between ports {self.camera_A.port} and {self.camera_B.port}")
            points_A_undistorted = undistort(xy_A, self.camera_A)
            points_B_undistorted = undistort(xy_B, self.camera_B)

            # triangulate points outputs data in 4D homogenous coordinate system
            # note that these are in a world frame of reference
            xyzw_h = cv2.triangulatePoints(self.proj_A, self.proj_B, points_A_undistorted, points_B_undistorted)

            xyz_h = xyzw_h.T[:, :3]
            w = xyzw_h[3, :]
            xyz = np.divide(xyz_h.T, w).T  # convert to euclidean coordinates
        else:
            xyz = np.array([])

        # update the paired point packet with the 3d positions
        paired_points.xyz = xyz



# the PointHistory object requires a highly particular format
# all of the data is ultimately embedded in the initial camera array configuration
# and the calibration point data. These functions transform those two
# things into a PointHistory object that can be used to optimize the CaptureVolume
from pathlib import Path

import numpy as np
import pandas as pd

import caliscope.logger
from caliscope.cameras.camera_array import CameraArray
from caliscope.packets import FramePacket, PointPacket, SyncPacket
from caliscope.triangulate.array_stereo_triangulator import ArrayStereoTriangulator
from caliscope.triangulate.stereo_points_builder import (
    StereoPointsBuilder,
    StereoPointsPacket,
    SynchedStereoPointsPacket,
)

logger = caliscope.logger.get(__name__)


def get_stereotriangulated_table_original(camera_array: CameraArray, point_data_path: Path) -> pd.DataFrame:
    """
    Creates a table of stereotriangulated points from 2D point data stored in a CSV file.

    This function:
    1. Loads 2D point data from a CSV file
    2. Processes all sync indices to find point correspondences between camera pairs
    3. Triangulates corresponding points to get 3D coordinates
    4. Returns a DataFrame with the triangulated points data

    Args:
        camera_array: Contains camera parameters used for triangulation
        point_data_path: Path to the CSV file containing 2D point data

    Returns:
        pd.DataFrame: Table containing triangulated 3D points
    """
    logger.info(f"Beginning to create stereotriangulated points from data stored at {point_data_path}")
    point_data = pd.read_csv(point_data_path)

    xy_sync_indices = point_data["sync_index"].to_numpy()
    sync_indices = np.unique(xy_sync_indices)

    ports = [key for key in camera_array.port_index.keys()]

    # Pre-group data by sync_index to avoid repeated filtering operations
    # This creates a dictionary where keys are sync_indices and values are DataFrame subsets
    # Much more efficient than repeatedly filtering with point_data.query() in the loop
    sync_groups = {}
    for sync_index in sync_indices:
        # Create boolean mask for this sync_index (faster than .loc or .query)
        mask = point_data["sync_index"] == sync_index
        # Use the mask to get data for this sync_index
        sync_groups[sync_index] = point_data[mask]

    paired_point_builder = StereoPointsBuilder(ports)

    # Create the infrastructure for the pairwise triangulation
    array_triangulator = ArrayStereoTriangulator(camera_array)
    stereotriangulated_table = None

    # Calculate total number of sync indices for progress reporting
    total_sync_indices = len(sync_indices)

    logger.info("Begin reconstructing SyncPackets and SynchedStereoPairs... ")

    for i, sync_index in enumerate(sync_indices):
        # Report progress every 25 sync indices
        if i % 25 == 0:
            # More detailed progress report with percentage complete
            logger.info(
                            f"Processing stereotriangulation estimates...  {i}/{total_sync_indices} "
                            f"({i/total_sync_indices*100:.1f}%)"
                        )

        # Get pre-filtered data for this sync_index
        # This is faster than filtering the entire dataset each time
        port_points = sync_groups[sync_index]

        # initialize a dict to hold all the frame packets
        frame_packets = {}


        # Create frame packet for each port
        for port in ports:
            # Filter points for this port using boolean mask
            # This is faster than chained .loc operations
            port_mask = port_points["port"] == port
            port_data = port_points[port_mask]

            # Only create frame packet if there are points for this port
            if not port_data.empty:
                # Extract metadata - these are scalar values so iloc[0] is appropriate
                frame_time = port_data["frame_time"].iloc[0]
                frame_index = port_data["frame_index"].iloc[0]

                point_id = port_data["point_id"].to_numpy()

                img_loc_x = port_data["img_loc_x"].to_numpy()
                img_loc_y = port_data["img_loc_y"].to_numpy()
                img_loc = np.vstack([img_loc_x, img_loc_y]).T

                obj_loc_x = port_data["obj_loc_x"].to_numpy()
                obj_loc_y = port_data["obj_loc_y"].to_numpy()
                obj_loc = np.vstack([obj_loc_x, obj_loc_y]).T

                point_packet = PointPacket(point_id, img_loc, obj_loc)
                frame_packet = FramePacket(
                    port=port,
                    frame_index=frame_index,
                    frame_time=frame_time,
                    frame=None,
                    points=point_packet,
                )
                frame_packets[port] = frame_packet
            else:
                frame_packets[port] = None

        # create the sync packet for this sync index
        sync_packet = SyncPacket(sync_index, frame_packets)

        # get the paired point packets for all port pairs at this sync index
        synched_stereo_points: SynchedStereoPointsPacket = paired_point_builder.get_synched_paired_points(sync_packet)
        # print(synched_paired_points)
        array_triangulator.triangulate_synched_points(synched_stereo_points)

        for pair in synched_stereo_points.pairs:
            triangulated_pair: StereoPointsPacket = synched_stereo_points.stereo_points_packets[pair]
            if triangulated_pair is not None:
                if stereotriangulated_table is None:
                    stereotriangulated_table = triangulated_pair.to_table()
                else:
                    new_table = triangulated_pair.to_table()
                    for key, value in new_table.items():
                        stereotriangulated_table[key].extend(value)

    logger.info(f"Saving stereotriangulated_points.csv to {point_data_path.parent} for inspection")
    stereotriangulated_table = pd.DataFrame(stereotriangulated_table)
    stereotriangulated_table.to_csv(Path(point_data_path.parent, "stereotriangulated_points.csv"))

    logger.info("Returning dataframe of stereotriangulated points to caller")

    return stereotriangulated_table


def get_stereotriangulated_table(camera_array: CameraArray, point_data_path: Path) -> pd.DataFrame:
    """
    Creates a table of stereotriangulated points from 2D point data stored in a CSV file.
    - processing all sync indices at once.

    This function:
    1. Loads 2D point data from a CSV file
    2. Processes all sync indices to find point correspondences between camera pairs
    3. Triangulates corresponding points to get 3D coordinates
    4. Returns a DataFrame with the triangulated points data

    Args:
        camera_array: Contains camera parameters used for triangulation
        point_data_path: Path to the CSV file containing 2D point data

    Returns:
        pd.DataFrame: Table containing triangulated 3D points
    """
    logger.info(f"Beginning to create stereotriangulated points from data stored at {point_data_path}")
    point_data = pd.read_csv(point_data_path)

    # Store original values
    point_data['original_sync_index'] = point_data['sync_index']
    point_data['original_point_id'] = point_data['point_id']

    # Create unique point IDs by combining sync_index and point_id
    logger.info("Converting point_data to mimic 'single sync index' format")
    max_point_id = point_data['point_id'].max()
    point_id_multiplier = max_point_id + 1
    point_data['point_id'] = point_data['sync_index'] * point_id_multiplier + point_data['point_id']

    # Set all sync indices to 0 (treating all points as from the same virtual frame)
    point_data['sync_index'] = 0

    ports = [key for key in camera_array.port_index.keys()]

    # Process all cameras and points at once
    frame_packets = {}
    for port in ports:
        logger.info(f"Creating data for port {port}")
        port_data = point_data[point_data['port'] == port]
        if not port_data.empty:
            # Use any frame time since we're combining all frames
            frame_time = port_data['frame_time'].iloc[0]
            frame_index = 0

            point_id = port_data['point_id'].to_numpy()
            img_loc_x = port_data['img_loc_x'].to_numpy()
            img_loc_y = port_data['img_loc_y'].to_numpy()
            img_loc = np.vstack([img_loc_x, img_loc_y]).T

            obj_loc_x = port_data['obj_loc_x'].to_numpy()
            obj_loc_y = port_data['obj_loc_y'].to_numpy()
            obj_loc = np.vstack([obj_loc_x, obj_loc_y]).T

            point_packet = PointPacket(point_id, img_loc, obj_loc)
            frame_packet = FramePacket(
                port=port,
                frame_index=frame_index,
                frame_time=frame_time,
                frame=None,
                points=point_packet,
            )
            frame_packets[port] = frame_packet
        else:
            frame_packets[port] = None

    # Single triangulation pass for all points
    sync_packet = SyncPacket(0, frame_packets)
    paired_point_builder = StereoPointsBuilder(ports)
    array_triangulator = ArrayStereoTriangulator(camera_array)

    synched_stereo_points = paired_point_builder.get_synched_paired_points(sync_packet)
    array_triangulator.triangulate_synched_points(synched_stereo_points)

    # Collect and process results
    stereotriangulated_table = None
    for pair in synched_stereo_points.pairs:
        triangulated_pair = synched_stereo_points.stereo_points_packets[pair]
        if triangulated_pair is not None:
            if stereotriangulated_table is None:
                stereotriangulated_table = triangulated_pair.to_table()
            else:
                new_table = triangulated_pair.to_table()
                for key, value in new_table.items():
                    stereotriangulated_table[key].extend(value)

    # Convert to DataFrame and restore original values
    stereotriangulated_table = pd.DataFrame(stereotriangulated_table)

    # Decompose combined point_id back to original sync_index and point_id
    stereotriangulated_table['original_sync_index'] = stereotriangulated_table['point_id'] // point_id_multiplier
    stereotriangulated_table['original_point_id'] = stereotriangulated_table['point_id'] % point_id_multiplier

    # Update to original values
    stereotriangulated_table['sync_index'] = stereotriangulated_table['original_sync_index']
    stereotriangulated_table['point_id'] = stereotriangulated_table['original_point_id']
    stereotriangulated_table.drop(['original_sync_index', 'original_point_id'], axis=1, inplace=True)

    logger.info(f"Saving stereotriangulated_points.csv to {point_data_path.parent}")
    stereotriangulated_table.to_csv(Path(point_data_path.parent, "stereotriangulated_points.csv"))

    return stereotriangulated_table



from pathlib import Path
from queue import Queue
from threading import Event, Thread

import numpy as np
import pandas as pd

import caliscope.logger
from caliscope.cameras.camera_array import CameraArray
from caliscope.cameras.synchronizer import Synchronizer, SyncPacket
from caliscope.packets import XYZPacket
from caliscope.triangulate.triangulation import triangulate_sync_index

logger = caliscope.logger.get(__name__)


class SyncPacketTriangulator:
    """
    Will place 3d packets on subscribed queues and save consolidated data in csv
    format to output_path if provided
    """

    def __init__(
        self,
        camera_array: CameraArray,
        synchronizer: Synchronizer,
        recording_directory: Path = None,
        tracker_name: str = None,  # used only for getting the point names and tracker name
    ):
        self.camera_array = camera_array
        self.synchronizer = synchronizer
        self.recording_directory = recording_directory

        self.stop_thread = Event()
        self.stop_thread.clear()

        self.tracker_name = tracker_name

        self.xyz_history = {
            "sync_index": [],
            "point_id": [],
            "x_coord": [],
            "y_coord": [],
            "z_coord": [],
        }

        self.sync_packet_in_q = Queue(-1)
        self.synchronizer.subscribe_to_sync_packets(self.sync_packet_in_q)

        self.projection_matrices = self.camera_array.projection_matrices

        self.subscribers = []
        self.running = True
        self.thread = Thread(target=self.process_incoming, args=(), daemon=True)
        self.thread.start()

    def subscribe(self, queue: Queue):
        self.subscribers.append(queue)

    def unsubscriber(self, queue: Queue):
        self.subscribers.remove(queue)

    def process_incoming(self):
        # waiting to set running property here was causing issues with identifying state of thread.
        # set property to true then start thread...

        # self.running = True
        while not self.stop_thread.is_set():
            sync_packet: SyncPacket = self.sync_packet_in_q.get()

            if sync_packet is None:
                # No more sync packets after this... wind down
                self.stop_thread.set()
                logger.info("End processing of incoming sync packets...end signaled with `None` packet")
            else:
                logger.debug(
                    f"Sync Packet {sync_packet.sync_index} acquired with {sync_packet.frame_packet_count} frames"
                )
                # only attempt to process if data exists
                if sync_packet.frame_packet_count >= 2:
                    cameras, point_ids, imgs_xy = sync_packet.triangulation_inputs
                    logger.debug("Attempting to triangulate synced frames")
                    # prepare for jit
                    cameras = np.array(cameras)
                    point_ids = np.array(point_ids)
                    imgs_xy = np.array(imgs_xy)

                    logger.debug(f"Cameras are {cameras} and point_ids are {point_ids}")
                    if len(np.unique(cameras)) >= 2:
                        logger.debug(f"Points observed on cameras {np.unique(cameras)}")
                        point_id_xyz, points_xyz = triangulate_sync_index(
                            self.projection_matrices, cameras, point_ids, imgs_xy
                        )

                        logger.debug(
                            f"Synch Packet {sync_packet.sync_index} | Point ID: {point_id_xyz} | xyz: {points_xyz}"
                        )

                        xyz_packet = XYZPacket(sync_packet.sync_index, point_id_xyz, points_xyz)
                        logger.info(
                            f"Placing xyz pacKet for index {sync_packet.sync_index} with {len(xyz_packet.point_ids)} points"  # noqa E501
                        )
                        for q in self.subscribers:
                            q.put(xyz_packet)

                        # if self.output_path is not None:
                        self.add_packet_to_history(xyz_packet)

        self.running = False

        if self.recording_directory is not None:
            logger.info(f"Saving xyz point data to {self.recording_directory}")
            self.save_history()

    def add_packet_to_history(self, xyz_packet: XYZPacket):
        point_count = len(xyz_packet.point_ids)

        if point_count > 0:
            self.xyz_history["sync_index"].extend([xyz_packet.sync_index] * point_count)
            self.xyz_history["point_id"].extend(xyz_packet.point_ids)

            xyz_array = np.array(xyz_packet.point_xyz)
            self.xyz_history["x_coord"].extend(xyz_array[:, 0].tolist())
            self.xyz_history["y_coord"].extend(xyz_array[:, 1].tolist())
            self.xyz_history["z_coord"].extend(xyz_array[:, 2].tolist())

    def save_history(self) -> None:
        """
        If a recording directory is provided, then save the xyz directory into it
        If a tracker name is provided, then base name on the tracker name
        """
        df_xyz: pd.DataFrame = pd.DataFrame(self.xyz_history)

        if self.recording_directory is not None:
            if self.tracker_name is None:
                filename = "xyz.csv"
            else:
                filename = f"xyz_{self.tracker_name}.csv"
                df_xyz.to_csv(Path(self.recording_directory, filename))



from dataclasses import dataclass, field
from pathlib import Path

import numpy as np
import pandas as pd

from caliscope.packets import XYZPacket
from caliscope.trackers.tracker_enum import TrackerEnum


@dataclass
class MotionTrial:
    """
    Motion trial loaded in from output csv
    """

    xyz_csv: Path
    xyz_packets: dict = field(default_factory=dict[int:XYZPacket])

    def __post_init__(self):
        # assert(isinstance(self.xyz_csv,Path))
        # assert(self.xyz_csv.exists())

        tracker_name = self.xyz_csv.stem[4:]  # peel off "xyz_"
        self.tracker = TrackerEnum[tracker_name].value()

        if hasattr(self.tracker, "wireframe"):
            self.wireframe = self.tracker.wireframe
        else:
            self.wireframe = None

        self.xyz_df = pd.read_csv(self.xyz_csv, engine="pyarrow")
        sync_indices = self.xyz_df["sync_index"].unique()

        self.start_index = sync_indices.min()
        self.end_index = sync_indices.max()
        self.xyz_packets = {}

        if len(sync_indices) == 0:
            self.is_empty = True
        else:
            self.is_empty = False

    def get_xyz(self, sync_index: int) -> XYZPacket:
        """
        Cache packets as they are initially read off
        """
        if sync_index not in self.xyz_packets:
            current_sync_index = self.xyz_df["sync_index"] == sync_index
            point_ids = self.xyz_df["point_id"][current_sync_index]

            x = self.xyz_df["x_coord"][current_sync_index]
            y = self.xyz_df["y_coord"][current_sync_index]
            z = self.xyz_df["z_coord"][current_sync_index]

            xyz = np.column_stack([x, y, z])
            self.xyz_packets[sync_index] = XYZPacket(sync_index=sync_index, point_ids=point_ids, point_xyz=xyz)

        return self.xyz_packets[sync_index]

    def update_wireframe(self, sync_index: int):
        xyz_packet = self.get_xyz(sync_index)
        self.wireframe.set_points(xyz_packet)



import random
from queue import Queue
from threading import Event, Thread

import cv2

import caliscope.logger
from caliscope.cameras.camera_array import CameraData
from caliscope.packets import FramePacket
from caliscope.recording.recorded_stream import RecordedStream

logger = caliscope.logger.get(__name__)


class IntrinsicCalibrator:
    """
    Takes a recorded stream and determines a CameraData object from it
    Stream needs to have a charuco tracker assigned to it
    """

    def __init__(self, camera_data: CameraData, stream: RecordedStream):
        self.camera = camera_data  # pointer needed to update params
        self.stream = stream
        self.initialize_point_history()

        self.frame_packet_q = Queue()
        self.stream.subscribe(self.frame_packet_q)

        # The following group of parameters relate to the autopopulation of the calibrator
        self.grid_history_q = Queue()  # for passing ids, img_loc used in calibration
        self.auto_store_data = Event()
        self.auto_store_data.clear()
        self.auto_pop_frame_wait = (
            0  # how many frames will you hold ofF checking if the board can be added to the calibration pile.
        )
        self.target_grid_count = 0

        self.harvest_frames()

    def harvest_frames(self):
        self.stop_event = Event()
        self.stop_event.clear()

        def harvest_worker():
            while True:
                frame_packet = self.frame_packet_q.get()
                if self.stop_event.is_set():
                    break

                self.add_frame_packet(frame_packet)

            logger.info(f"Harvest frames successfully ended in calibrator for port {self.stream.port}")

        self.harvest_thread = Thread(target=harvest_worker, args=[], daemon=True)
        self.harvest_thread.start()

    def stop(self):
        logger.info("Beginning to stop intrinsic calibrator")
        self.stop_event.set()
        self.stream.unsubscribe(self.frame_packet_q)
        self.frame_packet_q.put(-1)

    @property
    def grid_count(self):
        """How many sets of corners have been collected up to this point"""
        return len(self.calibration_frame_indices)

    @property
    def image_size(self):
        image_size = list(self.camera.size)
        image_size.reverse()  # for some reason...
        image_size.append(3)

        return image_size

    def initialize_point_history(self):
        # list of frame_indices that will go into the calibration
        self.calibration_frame_indices = []

        # dictionaries here are indexed by frame_index
        self.all_ids = {}
        self.all_img_loc = {}
        self.all_obj_loc = {}

    def add_frame_packet(self, frame_packet: FramePacket):
        """
        Point data from frame packet is stored, indexed by the frame index
        """
        index = frame_packet.frame_index

        if index != -1:  # indicates end of stream
            self.all_ids[index] = frame_packet.points.point_id
            self.all_img_loc[index] = frame_packet.points.img_loc
            self.all_obj_loc[index] = frame_packet.points.obj_loc

            self.active_frame_index = index

            # when auto store data is set, the stream should be pushing out all
            # frames consecutively from the beginning
            if self.auto_store_data.is_set():
                point_id = frame_packet.points.point_id

                if point_id.size == 0:
                    corner_count = 0
                else:
                    corner_count = frame_packet.points.point_id.shape[0]
                logger.debug(f"Corner count is {corner_count} and frame wait is {self.auto_pop_frame_wait}")
                if self.auto_pop_frame_wait == 0 and corner_count >= self.threshold_corner_count:
                    # add frame to calibration data and reset the wait time
                    self.add_calibration_frame_index(index)
                    self.auto_pop_frame_wait = self.wait_between
                else:
                    # count down to the next frame to consider autopopulating
                    self.auto_pop_frame_wait = max(self.auto_pop_frame_wait - 1, 0)

                logger.debug(f"Current index is {index}")
                if index == self.stream.last_frame_index:
                    # end of stream, so stop auto pop and backfill to hit grid target
                    logger.info("End of autopop detected...")
                    self.auto_store_data.clear()
                    self.backfill_calibration_frames()

    def backfill_calibration_frames(self):
        logger.info(
            f"Initiating backfill of frames to hit target grid count of {self.target_grid_count}...currently at {self.grid_count}"  # noqa: E501
        )
        actual_grid_count = len(self.calibration_frame_indices)
        logger.info(f"Actual Grid Count: {actual_grid_count}")
        # build new frame list
        new_potential_frames = []
        for frame_index, ids in self.all_ids.items():
            if frame_index not in self.calibration_frame_indices:
                if len(ids) >= 6:  # believe this may be a requirement of the calibration algorithm
                    new_potential_frames.append(frame_index)

        sample_size = self.target_grid_count - actual_grid_count
        sample_size = min(sample_size, len(new_potential_frames))
        sample_size = max(sample_size, 0)

        random_frames = random.sample(new_potential_frames, sample_size)
        for frame in random_frames:
            self.add_calibration_frame_index(frame)

    def add_calibration_frame_index(self, frame_index: int):
        """
        A "side effect" of this method is that the corner id and img_loc
        data is placed on a q. This q is consumed by the frame_emitter
        which will update the grid capture history based with those grids

        This allows the GUI element (frame emitter) to stay in sync with the
        calibrator.
        """
        logger.info(f"Adding frame data to calibration inputs for frame index {frame_index}")
        self.calibration_frame_indices.append(frame_index)

        # Backchannel communication to frame_emitter to keep things aligned
        ids = self.all_ids[frame_index]
        img_loc = self.all_img_loc[frame_index]
        self.grid_history_q.put((ids, img_loc))

    def clear_calibration_data(self):
        logger.info("Clearing calibration data..")
        self.calibration_frame_indices = []
        self.set_calibration_inputs()

    def initiate_auto_pop(self, wait_between, threshold_corner_count, target_grid_count):
        """
        This will enable actions within self.add_frame_packet

        Now when frame_packets are read in from the stream, the

        """
        logger.info(f"Initiating autopopulation of corner data in port {self.camera.port}")
        self.clear_calibration_data()
        self.wait_between = wait_between
        self.threshold_corner_count = threshold_corner_count
        self.target_grid_count = target_grid_count
        self.initialize_point_history()
        self.auto_store_data.set()

    def set_calibration_inputs(self):
        """
        The data that will ultimately go into the calibration is determined by the calibration
        frame indices. This is where the calibration data is populated based on which frames
        have been flagged (either by the user via self.add_calibration_frame_index or by autopopulated
        within self.add_frame_packet when autopop is enabled.)
        """
        self.calibration_point_ids = []
        self.calibration_img_loc = []
        self.calibration_obj_loc = []
        logger.info(f"Blank calibration inputs initialized at port {self.camera.port }")
        for index in self.calibration_frame_indices:
            id_count = len(self.all_ids[index])
            if id_count > 3:  # I believe this is a requirement of opencv
                self.calibration_point_ids.append(self.all_ids[index])
                self.calibration_img_loc.append(self.all_img_loc[index])
                self.calibration_obj_loc.append(self.all_obj_loc[index])
            else:
                logger.info(
                    f"Note that empty data stored in frame index {index}. This is not being used in the calibration"
                )
        logger.info(f"Total size of inputs is {len(self.calibration_point_ids)}")

    def calibrate_camera(self):
        """
        Use the recorded image corner positions along with the objective
        corner positions based on the board definition to calculated
        the camera matrix and distortion parameters
        """
        self.set_calibration_inputs()

        logger.info(f"Calibrating camera {self.camera.port}....")

        width = self.stream.size[0]
        height = self.stream.size[1]

        self.error, self.mtx, self.dist, self.rvecs, self.tvecs = cv2.calibrateCamera(
            self.calibration_obj_loc,
            self.calibration_img_loc,
            (width, height),
            None,
            None,
        )

        # fix extra dimension in return value of cv2.calibrateCamera
        self.dist = self.dist[0]

        self.update_camera()
        self.is_calibrated = True

        logger.info(f"Error: {self.error}")
        logger.info(f"Camera Matrix: {self.mtx}")
        logger.info(f"Distortion: {self.dist}")
        logger.info(f"Grid Count: {self.grid_count}")

    def update_camera(self):
        logger.info(f"Setting calibration params on camera {self.camera.port}")

        # ret is RMSE of reprojection
        self.camera.error = round(self.error, 3)
        self.camera.matrix = self.mtx
        self.camera.distortions = self.dist
        self.camera.grid_count = self.grid_count



from queue import Queue
from threading import Thread

import cv2
import mediapipe as mp
import numpy as np

import caliscope.logger

# cap = cv2.VideoCapture(0)
from caliscope.packets import PointPacket
from caliscope.tracker import Tracker
from caliscope.trackers.helper import apply_rotation, unrotate_points

logger = caliscope.logger.get(__name__)


class FaceTracker(Tracker):
    # Initialize MediaPipe Facemeshes and Drawing utility
    def __init__(self) -> None:
        self.in_queue = Queue(-1)
        self.out_queue = Queue(-1)

        # each port gets its own mediapipe context manager
        # use a dictionary of queues for passing
        self.in_queues = {}
        self.out_queues = {}
        self.threads = {}

    @property
    def name(self):
        return "FACE"

    def run_frame_processor(self, port: int, rotation_count: int):
        # Create a MediaPipe FaceMesh instance
        with mp.solutions.face_mesh.FaceMesh(
            static_image_mode=False, max_num_faces=1, refine_landmarks=True, min_detection_confidence=0.5
        ) as facemeshes:
            while True:
                frame = self.in_queues[port].get()
                # apply rotation as needed
                frame = apply_rotation(frame, rotation_count)

                height, width, color = frame.shape
                # Convert the image to RGB format
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                results = facemeshes.process(frame)

                # initialize variables so none will be created if no points detected
                point_ids = []
                landmark_xy = []

                if results.multi_face_landmarks:
                    for face_landmarks in results.multi_face_landmarks:
                        for landmark_id, landmark in enumerate(face_landmarks.landmark):
                            point_ids.append(landmark_id)

                            # mediapipe expresses in terms of percent of frame, so must map to pixel position
                            x, y = int(landmark.x * width), int(landmark.y * height)
                            landmark_xy.append((x, y))

                point_ids = np.array(point_ids)
                landmark_xy = np.array(landmark_xy)
                landmark_xy = unrotate_points(landmark_xy, rotation_count, width, height)

                point_packet = PointPacket(point_ids, landmark_xy)

                self.out_queues[port].put(point_packet)

    def get_points(self, frame: np.ndarray, port: int, rotation_count: int) -> PointPacket:
        if port not in self.in_queues.keys():
            self.in_queues[port] = Queue(1)
            self.out_queues[port] = Queue(1)

            self.threads[port] = Thread(
                target=self.run_frame_processor,
                args=(port, rotation_count),
                daemon=True,
            )
            self.threads[port].start()

        self.in_queues[port].put(frame)
        point_packet = self.out_queues[port].get()

        return point_packet

    def get_point_name(self, point_id: int) -> str:
        return str(point_id)

    def scatter_draw_instructions(self, point_id: int) -> dict:
        if POINT_ID2NAME[point_id] == "silhouette":
            rules = {"radius": 5, "color": (0, 0, 0), "thickness": 3}
        if POINT_ID2NAME[point_id] in ("lipsUpperOuter", "lipsLowerOuter", "lipsUpperInner", "lipsLowerInner"):
            rules = {"radius": 5, "color": (128, 0, 0), "thickness": 3}
        if POINT_ID2NAME[point_id] in (
            "rightEyeUpper0",
            "rightEyeLower0",
            "rightEyeUpper1",
            "rightEyeLower1",
            "rightEyeUpper2",
            "rightEyeLower2",
            "rightEyeLower3",
            "rightEyeIris",
        ):
            rules = {"radius": 5, "color": (0, 255, 0), "thickness": 3}
        if POINT_ID2NAME[point_id] in ("rightEyebrowUpper", "rightEyebrowLower"):
            rules = {"radius": 5, "color": (0, 128, 0), "thickness": 3}
        if POINT_ID2NAME[point_id] in (
            "leftEyeUpper0",
            "leftEyeLower0",
            "leftEyeUpper1",
            "leftEyeLower1",
            "leftEyeUpper2",
            "leftEyeLower2",
            "leftEyeLower3",
            "leftEyeIris",
        ):
            rules = {"radius": 5, "color": (255, 0, 0), "thickness": 3}
        if POINT_ID2NAME[point_id] in ("leftEyebrowUpper", "leftEyebrowLower"):
            rules = {"radius": 5, "color": (128, 0, 0), "thickness": 3}
        if POINT_ID2NAME[point_id] == "midwayBetweenEyes":
            rules = {"radius": 5, "color": (0, 0, 0), "thickness": 3}
        if POINT_ID2NAME[point_id] == (
            "noseTip",
            "noseBottom",
            "noseRightCorner",
            "noseLeftCorner",
            "leftCheek",
            "rightCheek",
        ):
            rules = {"radius": 5, "color": (0, 0, 255), "thickness": 3}
        else:
            rules = {"radius": 5, "color": (0, 0, 0), "thickness": 3}
        return rules


# Keypoint names copied from tfjs under Apache License 2.0, "AS-IS".
NAME2KEYPOINTS = {
    "silhouette": [
        10,
        338,
        297,
        332,
        284,
        251,
        389,
        356,
        454,
        323,
        361,
        288,
        397,
        365,
        379,
        378,
        400,
        377,
        152,
        148,
        176,
        149,
        150,
        136,
        172,
        58,
        132,
        93,
        234,
        127,
        162,
        21,
        54,
        103,
        67,
        109,
    ],
    "lipsUpperOuter": [61, 185, 40, 39, 37, 0, 267, 269, 270, 409, 291],
    "lipsLowerOuter": [146, 91, 181, 84, 17, 314, 405, 321, 375, 291],
    "lipsUpperInner": [78, 191, 80, 81, 82, 13, 312, 311, 310, 415, 308],
    "lipsLowerInner": [78, 95, 88, 178, 87, 14, 317, 402, 318, 324, 308],
    "rightEyeUpper0": [246, 161, 160, 159, 158, 157, 173],
    "rightEyeLower0": [33, 7, 163, 144, 145, 153, 154, 155, 133],
    "rightEyeUpper1": [247, 30, 29, 27, 28, 56, 190],
    "rightEyeLower1": [130, 25, 110, 24, 23, 22, 26, 112, 243],
    "rightEyeUpper2": [113, 225, 224, 223, 222, 221, 189],
    "rightEyeLower2": [226, 31, 228, 229, 230, 231, 232, 233, 244],
    "rightEyeLower3": [143, 111, 117, 118, 119, 120, 121, 128, 245],
    "rightEyebrowUpper": [156, 70, 63, 105, 66, 107, 55, 193],
    "rightEyebrowLower": [35, 124, 46, 53, 52, 65],
    "rightEyeIris": [473, 474, 475, 476, 477],
    "leftEyeUpper0": [466, 388, 387, 386, 385, 384, 398],
    "leftEyeLower0": [263, 249, 390, 373, 374, 380, 381, 382, 362],
    "leftEyeUpper1": [467, 260, 259, 257, 258, 286, 414],
    "leftEyeLower1": [359, 255, 339, 254, 253, 252, 256, 341, 463],
    "leftEyeUpper2": [342, 445, 444, 443, 442, 441, 413],
    "leftEyeLower2": [446, 261, 448, 449, 450, 451, 452, 453, 464],
    "leftEyeLower3": [372, 340, 346, 347, 348, 349, 350, 357, 465],
    "leftEyebrowUpper": [383, 300, 293, 334, 296, 336, 285, 417],
    "leftEyebrowLower": [265, 353, 276, 283, 282, 295],
    "leftEyeIris": [468, 469, 470, 471, 472],
    "midwayBetweenEyes": [168],
    "noseTip": [1],
    "noseBottom": [2],
    "noseRightCorner": [98],
    "noseLeftCorner": [327],
    "rightCheek": [205],
    "leftCheek": [425],
}

NAMES = sorted(NAME2KEYPOINTS)
POINT_ID2NAME = [None for _ in range(478)]
for i_n, name in enumerate(NAMES):
    for id in NAME2KEYPOINTS[name]:
        POINT_ID2NAME[id] = name



# There may be a mixed functionality here...I'm not sure. Between the corner
# detector and the corner drawer...like, there will need to be something that
# accumulates a frame of corners to be drawn onto the displayed frame.

import cv2
import numpy as np

import caliscope.calibration.draw_charuco
import caliscope.logger
from caliscope.packets import PointPacket
from caliscope.tracker import Tracker

logger = caliscope.logger.get(__name__)


class CharucoTracker(Tracker):
    def __init__(self, charuco):
        # need camera to know resolution and to assign calibration parameters
        # to camera
        self.charuco = charuco
        self.board = charuco.board
        self.dictionary_object = self.charuco.dictionary_object

        # for subpixel corner correction
        self.criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.0001)
        self.conv_size = (11, 11)  # Don't make this too large.

    @property
    def name(self):
        return "CHARUCO"

    def get_points(self, frame: np.ndarray, port: int, rotation_count: int) -> PointPacket:
        """Will check for charuco corners in the frame, if it doesn't find any,
        then it will look for corners in the mirror image of the frame"""

        # invert the frame for detection if needed
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)  # convert to gray
        if self.charuco.inverted:
            gray = ~gray  # invert

        ids, img_loc = self.find_corners_single_frame(gray, mirror=False)

        if not ids.any():
            gray = cv2.flip(gray, 1)
            ids, img_loc = self.find_corners_single_frame(gray, mirror=True)

        obj_loc = self.get_obj_loc(ids)
        point_packet = PointPacket(ids, img_loc, obj_loc)

        return point_packet

    def get_point_name(self, point_id: int) -> str:
        return str(point_id)

    def get_point_id(self, point_name: str) -> int:
        return int(point_name)

    def get_connected_points(self):
        return self.charuco.get_connected_points()

    def find_corners_single_frame(self, gray_frame, mirror):
        ids = np.array([])
        img_loc = np.array([])

        # detect if aruco markers are present
        aruco_corners, aruco_ids, rejected = cv2.aruco.detectMarkers(gray_frame, self.dictionary_object)

        # if so, then interpolate to the Charuco Corners and return what you found
        if len(aruco_corners) > 3:
            (
                success,
                _img_loc,
                _ids,
            ) = cv2.aruco.interpolateCornersCharuco(aruco_corners, aruco_ids, gray_frame, self.board)

            # This occasionally errors out...
            # only offers possible refinement so if it fails, just move along
            try:
                _img_loc = cv2.cornerSubPix(
                    gray_frame,
                    _img_loc,
                    self.conv_size,
                    (-1, -1),
                    self.criteria,
                )
            except Exception as e:
                logger.debug(f"Sub pixel detection failed: {e}")

            if success:
                # assign to tracker
                ids = _ids[:, 0]
                img_loc = _img_loc[:, 0]

                # flip coordinates if mirrored image fed in
                frame_width = gray_frame.shape[1]  # used for flipping mirrored corners back
                if mirror:
                    img_loc[:, 0] = frame_width - img_loc[:, 0]

        return ids, img_loc

    def get_obj_loc(self, ids: np.ndarray):
        """Objective position of charuco corners in a board frame of reference"""
        # if self.ids == np.array([0]):
        # print("wait")
        if len(ids) > 0:
            return self.board.getChessboardCorners()[ids, :]
        else:
            return np.array([])

    # @property
    def scatter_draw_instructions(self, point_id: int) -> dict:
        rules = {"radius": 5, "color": (0, 0, 220), "thickness": 3}
        return rules
